# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Compute the sum of two qubit registers using QFT."""

import numpy as np

from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit import QuantumRegister
from qiskit.circuit.library.basis_change import QFTGate


def adder_qft_d00(
    num_state_qubits: int, kind: str = "half", annotated: bool = False
) -> QuantumCircuit:
    r"""A circuit that uses QFT to perform in-place addition on two qubit registers.

    For registers with :math:`n` qubits, the QFT adder can perform addition modulo
    :math:`2^n` (with ``kind="fixed"``) or ordinary addition by adding a carry qubits (with
    ``kind="half"``). The fixed adder uses :math:`(3n^2 - n)/2` :class:`.CPhaseGate` operators,
    with an additional :math:`n` for the half adder.

    As an example, a non-fixed_point QFT adder circuit that performs addition on two 2-qubit sized
    registers is as follows:

    .. parsed-literal::

        a_0: ─────────■──────■────────■──────────────────────────────────
                      │      │        │
        a_1: ─────────┼──────┼────────┼────────■──────■──────────────────
             ┌──────┐ │      │        │P(π/4)  │      │P(π/2) ┌─────────┐
        b_0: ┤0     ├─┼──────┼────────■────────┼──────■───────┤0        ├
             │      │ │      │P(π/2)           │P(π)          │         │
        b_1: ┤1 Qft ├─┼──────■─────────────────■──────────────┤1 qft_dg ├
             │      │ │P(π)                                   │         │
       cout: ┤2     ├─■───────────────────────────────────────┤2        ├
             └──────┘                                         └─────────┘

    Args:
        num_state_qubits: The number of qubits in either input register for
            state :math:`|a\rangle` or :math:`|b\rangle`. The two input
            registers must have the same number of qubits.
        kind: The kind of adder, can be ``"half"`` for a half adder or
            ``"fixed"`` for a fixed-sized adder. A half adder contains a carry-out to represent
            the most-significant bit, but the fixed-sized adder doesn't and hence performs
            addition modulo ``2 ** num_state_qubits``.
        annotated: If ``True``, creates appropriate control and inverse operations as
            ``AnnotatedOperation`` objects.

    **References:**

    [1] T. G. Draper, Addition on a Quantum Computer, 2000.
    `arXiv:quant-ph/0008033 <https://arxiv.org/pdf/quant-ph/0008033.pdf>`_

    [2] Ruiz-Perez et al., Quantum arithmetic with the Quantum Fourier Transform, 2017.
    `arXiv:1411.5949 <https://arxiv.org/pdf/1411.5949.pdf>`_

    [3] Vedral et al., Quantum Networks for Elementary Arithmetic Operations, 1995.
    `arXiv:quant-ph/9511018 <https://arxiv.org/pdf/quant-ph/9511018.pdf>`_

    """

    if kind == "full":
        raise ValueError("The DraperQFTAdder only supports 'half' and 'fixed' as ``kind``.")

    if num_state_qubits < 1:
        raise ValueError("The number of qubits must be at least 1.")

    qr_a = QuantumRegister(num_state_qubits, name="a")
    qr_b = QuantumRegister(num_state_qubits, name="b")
    qr_list = [qr_a, qr_b]

    if kind == "half":
        qr_z = QuantumRegister(1, name="cout")
        qr_list.append(qr_z)

    # add registers
    circuit = QuantumCircuit(*qr_list)

    # define register containing the sum and number of qubits for QFT circuit
    qr_sum = qr_b[:] if kind == "fixed" else qr_b[:] + qr_z[:]
    num_sum = num_state_qubits if kind == "fixed" else num_state_qubits + 1

    # build QFT adder circuit
    qft = QFTGate(num_sum)
    circuit.append(qft, qr_sum[:])

    for j in range(num_state_qubits):
        for k in range(num_sum - j):
            lam = np.pi / (2**k)
            # note: if we were able to remove the final swaps from the QFTGate, we could
            # simply use qr_sum[(j + k)] here and avoid synthesizing two swap networks, which
            # can be elided and cancelled by the compiler
            circuit.cp(lam, qr_a[j], qr_sum[~(j + k)])

    circuit.append(qft.inverse(annotated=annotated), qr_sum[:])

    return circuit



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
==============================================
Random Circuits (:mod:`qiskit.circuit.random`)
==============================================

.. currentmodule:: qiskit.circuit.random

Overview
========

The :mod:`qiskit.circuit.random` module offers functions that can be used for generating 
arbitrary circuits with gates randomly selected from a given set of gates. 

These circuits can be used for benchmarking existing quantum hardware and estimating 
the performance of quantum circuit transpilers and software infrastructure.
The functions below can generate bespoke quantum circuits respecting various properties
such as number of qubits, depth of the circuit, coupling map, gate set, etc.

Generating arbitrary circuits
------------------------------------

.. autofunction:: random_circuit


Generating arbitrary circuits respecting qubit-coupling
--------------------------------------------------------------

.. autofunction:: random_circuit_from_graph


Generating arbitrary circuits with clifford gates
--------------------------------------------------------

.. autofunction:: random_clifford_circuit

"""

from .utils import random_circuit, random_clifford_circuit, random_circuit_from_graph



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Type-system definition for the expression tree."""


__all__ = [
    "Type",
    "Bool",
    "Duration",
    "Float",
    "Uint",
]

from qiskit._accelerate.circuit.classical.types import (
    Type,
    Bool,
    Uint,
    Float,
    Duration,
)  # pylint: disable=unused-import



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Two-qubit XX-YY gate."""

from __future__ import annotations

import math
from cmath import exp
from typing import Optional

import numpy as np

from qiskit.circuit.gate import Gate
from qiskit.circuit.parameterexpression import ParameterValueType, ParameterExpression
from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit._accelerate.circuit import StandardGate


class XXMinusYYGate(Gate):
    r"""XX-YY interaction gate.

    A 2-qubit parameterized XX-YY interaction. Its action is to induce
    a coherent rotation by some angle between :math:`|00\rangle` and :math:`|11\rangle`.

    **Circuit Symbol:**

    .. code-block:: text

             ┌───────────────┐
        q_0: ┤0              ├
             │  (XX-YY)(θ,β) │
        q_1: ┤1              ├
             └───────────────┘

    **Matrix Representation:**

    .. math::

        \newcommand{\rotationangle}{\frac{\theta}{2}}

        R_{XX-YY}(\theta, \beta) q_0, q_1 =
          RZ_1(\beta) \cdot \exp\left(-i \frac{\theta}{2} \frac{XX-YY}{2}\right) \cdot RZ_1(-\beta) =
            \begin{pmatrix}
                \cos\left(\rotationangle\right) & 0 & 0 & -i\sin\left(\rotationangle\right)e^{-i\beta} \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                -i\sin\left(\rotationangle\right)e^{i\beta} & 0 & 0 & \cos\left(\rotationangle\right)
            \end{pmatrix}
    """

    _standard_gate = StandardGate.XXMinusYY

    def __init__(
        self,
        theta: ParameterValueType,
        beta: ParameterValueType = 0,
        label: Optional[str] = "(XX-YY)",
    ):
        """Create new XX-YY gate.

        Args:
            theta: The rotation angle.
            beta: The phase angle.
            label: The label of the gate.
        """
        super().__init__("xx_minus_yy", 2, [theta, beta], label=label)

    def _define(self):
        """Default definition"""

        #       ┌─────┐  ┌────┐┌───┐     ┌─────────┐      ┌─────┐ ┌──────┐┌───┐
        # q_0: ─┤ Sdg ├──┤ √X ├┤ S ├──■──┤ Ry(θ/2) ├───■──┤ Sdg ├─┤ √Xdg ├┤ S ├
        #      ┌┴─────┴─┐├───┬┘└───┘┌─┴─┐├─────────┴┐┌─┴─┐├─────┤┌┴──────┤└───┘
        # q_1: ┤ Rz(-β) ├┤ S ├──────┤ X ├┤ Ry(-θ/2) ├┤ X ├┤ Sdg ├┤ Rz(β) ├─────
        #      └────────┘└───┘      └───┘└──────────┘└───┘└─────┘└───────┘

        self.definition = QuantumCircuit._from_circuit_data(
            StandardGate.XXMinusYY._get_definition(self.params), add_regs=True, name=self.name
        )

    def control(
        self,
        num_ctrl_qubits: int = 1,
        label: str | None = None,
        ctrl_state: str | int | None = None,
        annotated: bool | None = None,
    ):
        """Return a (multi-)controlled-(XX-YY) gate.

        Args:
            num_ctrl_qubits: number of control qubits.
            label: An optional label for the gate [Default: ``None``]
            ctrl_state: control state expressed as integer,
                string (e.g.``'110'``), or ``None``. If ``None``, use all 1s.
            annotated: indicates whether the controlled gate should be implemented
                as an annotated gate. If ``None``, this is set to ``True`` if
                the gate contains free parameters, in which case it cannot
                yet be synthesized.

        Returns:
            ControlledGate: controlled version of this gate.
        """
        if annotated is None:
            annotated = any(isinstance(p, ParameterExpression) for p in self.params)

        gate = super().control(
            num_ctrl_qubits=num_ctrl_qubits,
            label=label,
            ctrl_state=ctrl_state,
            annotated=annotated,
        )
        return gate

    def inverse(self, annotated: bool = False):
        """Inverse gate.

        Args:
            annotated: when set to ``True``, this is typically used to return an
                :class:`.AnnotatedOperation` with an inverse modifier set instead of a concrete
                :class:`.Gate`. However, for this class this argument is ignored as the inverse
                of this gate is always a :class:`.XXMinusYYGate` with inverse
                parameter values.

        Returns:
            XXMinusYYGate: inverse gate.
        """
        theta, beta = self.params
        return XXMinusYYGate(-theta, beta)

    def __array__(self, dtype=None, copy=None):
        """Gate matrix."""
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        theta, beta = self.params
        cos = math.cos(theta / 2)
        sin = math.sin(theta / 2)
        return np.array(
            [
                [cos, 0, 0, -1j * sin * exp(-1j * beta)],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [-1j * sin * exp(1j * beta), 0, 0, cos],
            ],
            dtype=dtype,
        )

    def power(self, exponent: float, annotated: bool = False):
        theta, beta = self.params
        return XXMinusYYGate(exponent * theta, beta)

    def __eq__(self, other):
        if isinstance(other, XXMinusYYGate):
            return self._compare_parameters(other)
        return False



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A circuit implementing a quadratic form on binary variables."""

from __future__ import annotations

from typing import Union, Optional, List
import math
from collections.abc import Sequence

import numpy as np

from qiskit.circuit import QuantumCircuit, QuantumRegister, ParameterExpression, Gate, CircuitError
from qiskit.utils.deprecation import deprecate_func
from ..basis_change import QFT, QFTGate

_ValueType = Union[int, float, np.integer, np.floating, ParameterExpression]


class QuadraticForm(QuantumCircuit):
    r"""Implements a quadratic form on binary variables encoded in qubit registers.

    A quadratic form on binary variables is a quadratic function :math:`Q` acting on a binary
    variable of :math:`n` bits, :math:`x = x_0 ... x_{n-1}`. For an integer matrix :math:`A`,
    an integer vector :math:`b` and an integer :math:`c` the function can be written as

    .. math::

        Q(x) = x^T A x + x^T b + c

    If :math:`A`, :math:`b` or :math:`c` contain scalar values, this circuit computes only
    an approximation of the quadratic form.

    Provided with :math:`m` qubits to encode the value, this circuit computes :math:`Q(x) \mod 2^m`
    in [two's complement](https://stackoverflow.com/questions/1049722/what-is-2s-complement)
    representation.

    .. math::

        |x\rangle_n |0\rangle_m \mapsto |x\rangle_n |(Q(x) + 2^m) \mod 2^m \rangle_m

    Since we use two's complement e.g. the value of :math:`Q(x) = 3` requires 2 bits to represent
    the value and 1 bit for the sign: `3 = '011'` where the first `0` indicates a positive value.
    On the other hand, :math:`Q(x) = -3` would be `-3 = '101'`, where the first `1` indicates
    a negative value and `01` is the two's complement of `3`.

    If the value of :math:`Q(x)` is too large to be represented with `m` qubits, the resulting
    bitstring is :math:`(Q(x) + 2^m) \mod 2^m)`.

    The implementation of this circuit is discussed in [1], Fig. 6.

    References:
        [1]: Gilliam et al., Grover Adaptive Search for Constrained Polynomial Binary Optimization.
             `arXiv:1912.04088 <https://arxiv.org/pdf/1912.04088.pdf>`_

    """

    @deprecate_func(
        since="2.1",
        additional_msg="Use the QuadraticFormGate instead.",
        removal_timeline="Qiskit 3.0",
    )
    def __init__(
        self,
        num_result_qubits: Optional[int] = None,
        quadratic: Optional[
            Union[np.ndarray, List[List[Union[float, ParameterExpression]]]]
        ] = None,
        linear: Optional[Union[np.ndarray, List[Union[float, ParameterExpression]]]] = None,
        offset: Optional[Union[float, ParameterExpression]] = None,
        little_endian: bool = True,
    ) -> None:
        r"""
        Args:
            num_result_qubits: The number of qubits to encode the result. Called :math:`m` in
                the class documentation.
            quadratic: A matrix containing the quadratic coefficients, :math:`A`.
            linear: An array containing the linear coefficients, :math:`b`.
            offset: A constant offset, :math:`c`.
            little_endian: Encode the result in little endianness.

        Raises:
            ValueError: If ``linear`` and ``quadratic`` have mismatching sizes.
            ValueError: If ``num_result_qubits`` is unspecified but cannot be determined because
                some values of the quadratic form are parameterized.
        """
        # check inputs match
        if quadratic is not None and linear is not None:
            if len(quadratic) != len(linear):
                raise ValueError("Mismatching sizes of quadratic and linear.")

        # temporarily set quadratic and linear to [] instead of None so we can iterate over them
        if quadratic is None:
            quadratic = []

        if linear is None:
            linear = []

        if offset is None:
            offset = 0

        num_input_qubits = np.max([1, len(linear), len(quadratic)])

        # deduce number of result bits if not added
        if num_result_qubits is None:
            # check no value is parameterized
            if (
                any(any(isinstance(q_ij, ParameterExpression) for q_ij in q_i) for q_i in quadratic)
                or any(isinstance(l_i, ParameterExpression) for l_i in linear)
                or isinstance(offset, ParameterExpression)
            ):
                raise ValueError(
                    "If the number of result qubits is not specified, the quadratic "
                    "form matrices/vectors/offset may not be parameterized."
                )
            num_result_qubits = self.required_result_qubits(quadratic, linear, offset)

        qr_input = QuantumRegister(num_input_qubits)
        qr_result = QuantumRegister(num_result_qubits)
        circuit = QuantumCircuit(qr_input, qr_result, name="Q(x)")

        # set quadratic and linear again to None if they were None
        if len(quadratic) == 0:
            quadratic = None

        if len(linear) == 0:
            linear = None

        scaling = np.pi * 2 ** (1 - num_result_qubits)

        # initial QFT (just hadamards)
        circuit.h(qr_result)

        if little_endian:
            qr_result = qr_result[::-1]

        # constant coefficient
        if offset != 0:
            for i, q_i in enumerate(qr_result):
                circuit.p(scaling * 2**i * offset, q_i)

        # the linear part consists of the vector and the diagonal of the
        # matrix, since x_i * x_i = x_i, as x_i is a binary variable
        for j in range(num_input_qubits):
            value = linear[j] if linear is not None else 0
            value += quadratic[j][j] if quadratic is not None else 0
            if value != 0:
                for i, q_i in enumerate(qr_result):
                    circuit.cp(scaling * 2**i * value, qr_input[j], q_i)

        # the quadratic part adds A_ij and A_ji as x_i x_j == x_j x_i
        if quadratic is not None:
            for j in range(num_input_qubits):
                for k in range(j + 1, num_input_qubits):
                    value = quadratic[j][k] + quadratic[k][j]
                    if value != 0:
                        for i, q_i in enumerate(qr_result):
                            circuit.mcp(scaling * 2**i * value, [qr_input[j], qr_input[k]], q_i)

        # add the inverse QFT
        iqft = QFT(num_result_qubits, do_swaps=False).inverse().reverse_bits()
        circuit.compose(iqft, qubits=qr_result[:], inplace=True)

        super().__init__(*circuit.qregs, name="Q(x)")
        self.compose(circuit.to_gate(), qubits=self.qubits, inplace=True)

    @staticmethod
    def required_result_qubits(
        quadratic: Union[np.ndarray, List[List[float]]],
        linear: Union[np.ndarray, List[float]],
        offset: float,
    ) -> int:
        """Get the number of required result qubits.

        Args:
            quadratic: A matrix containing the quadratic coefficients.
            linear: An array containing the linear coefficients.
            offset: A constant offset.

        Returns:
            The number of qubits needed to represent the value of the quadratic form
            in twos complement.
        """
        return QuadraticFormGate.required_result_qubits(quadratic, linear, offset)


class QuadraticFormGate(Gate):
    r"""Implements a quadratic form on binary variables encoded in qubit registers.

    A quadratic form on binary variables is a quadratic function :math:`Q` acting on a binary
    variable of :math:`n` bits, :math:`x = x_0 ... x_{n-1}`. For an integer matrix :math:`A`,
    an integer vector :math:`b` and an integer :math:`c` the function can be written as

    .. math::

        Q(x) = x^T A x + x^T b + c

    If :math:`A`, :math:`b` or :math:`c` contain scalar values, this circuit computes only
    an approximation of the quadratic form.

    Provided with :math:`m` qubits to encode the value, this circuit computes :math:`Q(x) \mod 2^m`
    in [two's complement](https://stackoverflow.com/questions/1049722/what-is-2s-complement)
    representation.

    .. math::

        |x\rangle_n |0\rangle_m \mapsto |x\rangle_n |(Q(x) + 2^m) \mod 2^m \rangle_m

    Since we use two's complement e.g. the value of :math:`Q(x) = 3` requires 2 bits to represent
    the value and 1 bit for the sign: `3 = '011'` where the first `0` indicates a positive value.
    On the other hand, :math:`Q(x) = -3` would be `-3 = '101'`, where the first `1` indicates
    a negative value and `01` is the two's complement of `3`.

    If the value of :math:`Q(x)` is too large to be represented with `m` qubits, the resulting
    bitstring is :math:`(Q(x) + 2^m) \mod 2^m)`.

    The implementation of this circuit is discussed in [1], Fig. 6.

    References:
        [1]: Gilliam et al., Grover Adaptive Search for Constrained Polynomial Binary Optimization.
             `arXiv:1912.04088 <https://arxiv.org/pdf/1912.04088.pdf>`_

    """

    def __init__(
        self,
        num_result_qubits: int | None = None,
        quadratic: Sequence[Sequence[float]] | None = None,
        linear: Sequence[Sequence[float]] | None = None,
        offset: float | None = None,
        label: str = "Q(x)",
    ):
        # check inputs match
        if quadratic is not None and linear is not None:
            if len(quadratic) != len(linear):
                raise ValueError("Mismatching sizes of quadratic and linear.")

        # temporarily set quadratic and linear to [] instead of None so we can iterate over them
        if quadratic is None:
            quadratic = []

        if linear is None:
            linear = []

        if offset is None:
            offset = 0

        self.num_input_qubits = np.max([1, len(linear), len(quadratic)])

        # deduce number of result bits if not added
        if num_result_qubits is None:
            num_result_qubits = self.required_result_qubits(quadratic, linear, offset)

        self.num_result_qubits = num_result_qubits
        self.quadratic = quadratic
        self.linear = linear
        self.offset = offset

        num_qubits = int(self.num_input_qubits + self.num_result_qubits)
        super().__init__("QuadraticForm", num_qubits, [], label=label)

    @staticmethod
    def required_result_qubits(
        quadratic: Sequence[Sequence[float]],
        linear: Sequence[float],
        offset: float,
    ) -> int:
        """Get the number of required result qubits.

        Args:
            quadratic: A matrix containing the quadratic coefficients.
            linear: An array containing the linear coefficients.
            offset: A constant offset.

        Returns:
            The number of qubits needed to represent the value of the quadratic form
            in twos complement.
        """

        bounds = []  # bounds = [minimum value, maximum value]
        for condition in [lambda x: x < 0, lambda x: x > 0]:
            bound = 0.0
            bound += sum(sum(q_ij for q_ij in q_i if condition(q_ij)) for q_i in quadratic)
            bound += sum(l_i for l_i in linear if condition(l_i))
            bound += offset if condition(offset) else 0
            bounds.append(bound)

        # the minimum number of qubits is the number of qubits needed to represent
        # the minimum/maximum value plus one sign qubit
        num_qubits_for_min = math.ceil(math.log2(max(-bounds[0], 1)))
        num_qubits_for_max = math.ceil(math.log2(bounds[1] + 1))
        num_result_qubits = 1 + max(num_qubits_for_min, num_qubits_for_max)

        return num_result_qubits

    def validate_parameter(self, parameter):
        if isinstance(parameter, _ValueType):
            return parameter

        if isinstance(parameter, (np.ndarray, Sequence)):
            if all(isinstance(el, _ValueType) for el in parameter):
                return parameter
            for params in parameter:
                if not all(isinstance(el, _ValueType) for el in params):
                    raise CircuitError(
                        f"Invalid parameter type {type(parameter)} for QuadraticFormGate"
                    )

            return parameter

        return super().validate_parameter(parameter)

    def _define(self):
        quadratic, linear, offset = self.quadratic, self.linear, self.offset

        qr_input = QuantumRegister(self.num_input_qubits)
        qr_result = QuantumRegister(self.num_result_qubits)
        circuit = QuantumCircuit(qr_input, qr_result)

        # set quadratic and linear again to None if they were None
        if len(quadratic) == 0:
            quadratic = None

        if len(linear) == 0:
            linear = None

        scaling = np.pi * 2 ** (1 - self.num_result_qubits)

        # initial QFT
        qft = QFTGate(self.num_result_qubits)
        circuit.append(qft, qr_result)

        # constant coefficient
        if offset != 0:
            for i, q_i in enumerate(qr_result):
                circuit.p(scaling * 2**i * offset, q_i)

        # the linear part consists of the vector and the diagonal of the
        # matrix, since x_i * x_i = x_i, as x_i is a binary variable
        for j in range(self.num_input_qubits):
            value = linear[j] if linear is not None else 0
            value += quadratic[j][j] if quadratic is not None else 0
            if value != 0:
                for i, q_i in enumerate(qr_result):
                    circuit.cp(scaling * 2**i * value, qr_input[j], q_i)

        # the quadratic part adds A_ij and A_ji as x_i x_j == x_j x_i
        if quadratic is not None:
            for j in range(self.num_input_qubits):
                for k in range(j + 1, self.num_input_qubits):
                    value = quadratic[j][k] + quadratic[k][j]
                    if value != 0:
                        for i, q_i in enumerate(qr_result):
                            circuit.mcp(scaling * 2**i * value, [qr_input[j], qr_input[k]], q_i)

        # add the inverse QFT
        iqft = qft.inverse()
        circuit.append(iqft, qr_result)

        self.definition = circuit



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Test random circuit generation utility."""
from collections import defaultdict
import rustworkx as rx
import numpy as np
import ddt
from qiskit.circuit import QuantumCircuit, ClassicalRegister, Clbit
from qiskit.circuit import Measure
from qiskit.circuit.exceptions import CircuitError
from qiskit.circuit.random import random_circuit
from qiskit.circuit.random.utils import random_circuit_from_graph
from qiskit.converters import circuit_to_dag
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestCircuitRandom(QiskitTestCase):
    """Testing qiskit.circuit.random"""

    def test_simple_random(self):
        """Test creating a simple random circuit."""
        circ = random_circuit(num_qubits=5, depth=4)
        self.assertIsInstance(circ, QuantumCircuit)
        self.assertEqual(circ.width(), 5)
        self.assertEqual(circ.depth(), 4)

    def test_random_depth_0(self):
        """Test random depth 0 circuit."""
        circ = random_circuit(num_qubits=1, depth=0)
        self.assertEqual(circ.width(), 1)
        self.assertEqual(circ.depth(), 0)

    def test_random_measure(self):
        """Test random circuit with final measurement."""
        num_qubits = depth = 3
        circ = random_circuit(num_qubits, depth, measure=True)
        self.assertEqual(circ.width(), 2 * num_qubits)
        dag = circuit_to_dag(circ)
        for nd in list(dag.topological_op_nodes())[-num_qubits:]:
            self.assertIsInstance(nd.op, Measure)

    def test_random_circuit_conditional_reset(self):
        """Test generating random circuits with conditional and reset."""
        num_qubits = 1
        depth = 100
        circ = random_circuit(num_qubits, depth, conditional=True, reset=True, seed=5)
        self.assertEqual(circ.width(), 2 * num_qubits)
        self.assertIn("reset", circ.count_ops())

    def test_large_conditional(self):
        """Test that conditions do not fail with large conditionals.  Regression test of gh-6994."""
        # The main test is that this call actually returns without raising an exception.
        circ = random_circuit(64, 2, conditional=True, seed=0)
        # Test that at least one instruction had a condition generated.  It's possible that this
        # fails due to very bad luck with the random seed - if so, change the seed to ensure that a
        # condition _is_ generated, because we need to test that generation doesn't error.
        conditions = (getattr(instruction.operation, "_condition", None) for instruction in circ)
        conditions = [x for x in conditions if x is not None]
        self.assertNotEqual(conditions, [])
        for register, value in conditions:
            self.assertIsInstance(register, (ClassicalRegister, Clbit))
            # Condition values always have to be Python bigints (of which `bool` is a subclass), not
            # any of Numpy's fixed-width types, for example.
            self.assertIsInstance(value, int)

    def test_random_mid_circuit_measure_conditional(self):
        """Test random circuit with mid-circuit measurements for conditionals."""
        num_qubits = depth = 2
        circ = random_circuit(num_qubits, depth, conditional=True, seed=16)
        self.assertEqual(circ.width(), 2 * num_qubits)
        op_names = [instruction.operation.name for instruction in circ]
        self.assertEqual(4, len(op_names))

        # Before a condition, there needs to be measurement in all the qubits.
        self.assertEqual(["measure"] * num_qubits, op_names[1 : 1 + num_qubits])
        conditions = [
            bool(getattr(instruction.operation, "condition", None)) for instruction in circ
        ]
        self.assertEqual([False, False, False, True], conditions)

    def test_random_circuit_num_operand_distribution(self):
        """Test that num_operand_distribution argument generates gates in correct proportion"""
        num_qubits = 50
        depth = 300
        num_op_dist = {2: 0.25, 3: 0.25, 1: 0.25, 4: 0.25}
        circ = random_circuit(
            num_qubits, depth, num_operand_distribution=num_op_dist, seed=123456789
        )
        total_gates = circ.size()
        self.assertEqual(circ.width(), num_qubits)
        self.assertEqual(circ.depth(), depth)
        gate_qubits = [instruction.operation.num_qubits for instruction in circ]
        gate_type_counter = np.bincount(gate_qubits, minlength=5)
        for gate_type, prob in sorted(num_op_dist.items()):
            self.assertAlmostEqual(prob, gate_type_counter[gate_type] / total_gates, delta=0.1)

    def test_random_circuit_2and3_qubit_gates_only(self):
        """
        Test that the generated random circuit only has 2 and 3 qubit gates,
        while disallowing 1-qubit and 4-qubit gates if
        num_operand_distribution = {2: some_prob, 3: some_prob}
        """
        num_qubits = 10
        depth = 200
        num_op_dist = {2: 0.5, 3: 0.5}
        circ = random_circuit(num_qubits, depth, num_operand_distribution=num_op_dist, seed=200)
        total_gates = circ.size()
        gate_qubits = [instruction.operation.num_qubits for instruction in circ]
        gate_type_counter = np.bincount(gate_qubits, minlength=5)
        # Testing that the distribution of 2 and 3 qubit gate matches with given distribution
        for gate_type, prob in sorted(num_op_dist.items()):
            self.assertAlmostEqual(prob, gate_type_counter[gate_type] / total_gates, delta=0.1)
        # Testing that there are no 1-qubit gate and 4-qubit in the generated random circuit
        self.assertEqual(gate_type_counter[1], 0.0)
        self.assertEqual(gate_type_counter[4], 0.0)

    def test_random_circuit_3and4_qubit_gates_only(self):
        """
        Test that the generated random circuit only has 3 and 4 qubit gates,
        while disallowing 1-qubit and 2-qubit gates if
        num_operand_distribution = {3: some_prob, 4: some_prob}
        """
        num_qubits = 10
        depth = 200
        num_op_dist = {3: 0.5, 4: 0.5}
        circ = random_circuit(
            num_qubits, depth, num_operand_distribution=num_op_dist, seed=11111111
        )
        total_gates = circ.size()
        gate_qubits = [instruction.operation.num_qubits for instruction in circ]
        gate_type_counter = np.bincount(gate_qubits, minlength=5)
        # Testing that the distribution of 3 and 4 qubit gate matches with given distribution
        for gate_type, prob in sorted(num_op_dist.items()):
            self.assertAlmostEqual(prob, gate_type_counter[gate_type] / total_gates, delta=0.1)
        # Testing that there are no 1-qubit gate and 2-qubit in the generated random circuit
        self.assertEqual(gate_type_counter[1], 0.0)
        self.assertEqual(gate_type_counter[2], 0.0)

    def test_random_circuit_with_zero_distribution(self):
        """
        Test that the generated random circuit only has 3 and 4 qubit gates,
        while disallowing 1-qubit and 2-qubit gates if
        num_operand_distribution = {1: 0.0, 2: 0.0, 3: some_prob, 4: some_prob}
        """
        num_qubits = 10
        depth = 200
        num_op_dist = {1: 0.0, 2: 0.0, 3: 0.5, 4: 0.5}
        circ = random_circuit(num_qubits, depth, num_operand_distribution=num_op_dist, seed=12)
        total_gates = circ.size()
        gate_qubits = [instruction.operation.num_qubits for instruction in circ]
        gate_type_counter = np.bincount(gate_qubits, minlength=5)
        # Testing that the distribution of 3 and 4 qubit gate matches with given distribution
        for gate_type, prob in sorted(num_op_dist.items()):
            self.assertAlmostEqual(prob, gate_type_counter[gate_type] / total_gates, delta=0.1)
        # Testing that there are no 1-qubit gate and 2-qubit in the generated random circuit
        self.assertEqual(gate_type_counter[1], 0.0)
        self.assertEqual(gate_type_counter[2], 0.0)


def incomplete_graph(n_nodes):
    # pylint: disable=missing-function-docstring
    pydi_graph = rx.generators.directed_complete_graph(n_nodes)
    pydi_graph.remove_edge(1, 3)
    return pydi_graph


def digraph_with_no_edges(n_nodes):
    # pylint: disable=missing-function-docstring
    graph = rx.PyDiGraph()
    graph.add_nodes_from(range(n_nodes))
    return graph


test_cases = (
    (rx.generators.directed_cycle_graph(5), 1550),
    (rx.generators.directed_mesh_graph(4), 87978),
    # The (almost) fully connected graph.
    (incomplete_graph(4), 154),
    (rx.generators.directed_heavy_hex_graph(3), 458),
    # Sparse connected graph
    (rx.generators.directed_path_graph(10), 458),
    # A graph with no edges, should yeild a circuit with no edges,
    # this means there would be no 2Q gates on that circuit.
    (digraph_with_no_edges(10), 0),
    # A list of tuples of control qubit, target qubit, and edge probability
    # is also acceptable.
    (
        [
            (0, 13, 21),
            (1, 13, 20),
            (1, 14, 15),
            (2, 14, 21),
            (3, 15, 10),
            (4, 15, 16),
            (4, 16, 21),
            (5, 16, 11),
            (6, 17, 11),
            (7, 17, 17),
            (7, 18, 15),
            (8, 18, 20),
            (0, 9, 13),
            (3, 9, 13),
            (5, 12, 22),
            (8, 12, 17),
            (10, 14, 11),
            (10, 16, 19),
            (11, 15, 12),
            (11, 17, 21),
        ],
        0,
    ),
)


@ddt.ddt
class TestRandomCircuitFromGraph(QiskitTestCase):
    """Testing random_circuit_from_graph from
    qiskit.circuit.random.utils.py"""

    @ddt.data(*test_cases)
    @ddt.unpack
    def test_simple_random(self, inter_graph, seed):
        """Test creating a simple random circuit."""
        n_nodes = 0
        if isinstance(inter_graph, list):
            for ctrl, trgt, _ in inter_graph:
                if ctrl > n_nodes:
                    n_nodes = ctrl
                if trgt > n_nodes:
                    n_nodes = trgt
            n_nodes += 1  # ctrl, trgt are qubit indices.
        else:
            n_nodes = inter_graph.num_nodes()

        circ = random_circuit_from_graph(interaction_graph=inter_graph, seed=seed)

        self.assertIsInstance(circ, QuantumCircuit)
        self.assertEqual(circ.width(), n_nodes)

    @ddt.data(*test_cases)
    @ddt.unpack
    def test_min_times_qubit_pair_usage(self, inter_graph, seed):
        """the `min_2q_gate_per_edge` parameter specifies how often each qubit-pair must at
        least be used in a two-qubit gate before the circuit is returned"""

        freq = 8  # Some arbitrary repetations, don't put 1.
        qc = random_circuit_from_graph(
            interaction_graph=inter_graph, min_2q_gate_per_edge=freq, seed=seed
        )
        count_register = defaultdict(int)

        for c_instr in qc._data:
            qubits = c_instr.qubits
            if len(qubits) == 2:
                count_register[(qubits[0]._index, qubits[1]._index)] += 1

        for occurrence in count_register.values():
            self.assertLessEqual(freq, occurrence)

    @ddt.data(*test_cases)
    @ddt.unpack
    def test_random_measure(self, inter_graph, seed):
        """Test random circuit with final measurement."""

        qc = random_circuit_from_graph(interaction_graph=inter_graph, measure=True, seed=seed)
        self.assertIn("measure", qc.count_ops())

    @ddt.data(*test_cases)
    @ddt.unpack
    def test_random_circuit_conditional_reset(self, inter_graph, seed):
        """Test generating random circuits with conditional and reset."""
        qc = random_circuit_from_graph(
            interaction_graph=inter_graph,
            min_2q_gate_per_edge=2,
            conditional=True,
            reset=True,
            seed=seed,  # Do not change the seed or the args.
            prob_conditional=0.95,
            prob_reset=0.50,
        )

        # Check if reset is applied.
        self.assertIn("reset", qc.count_ops())

        # Now, checking for conditionals
        cond_counter = 0
        for instr in qc:
            cond = getattr(instr.operation, "_condition", None)
            if not cond is None:
                cond_counter += 1
                break  # even one conditional is enough for the check.

        # See if conditionals are present.
        self.assertNotEqual(cond_counter, 0)

    @ddt.data(*test_cases)
    @ddt.unpack
    def test_2q_gates_applied_to_edges_from_interaction_graph(self, inter_graph, seed):
        """Test 2Q gates are applied to the qubit-pairs given by the interaction graph supplied"""
        qc = random_circuit_from_graph(
            interaction_graph=inter_graph,
            min_2q_gate_per_edge=2,
            measure=True,
            conditional=True,
            reset=True,
            seed=seed,  # Do not change the seed or args
            prob_conditional=0.41,
            prob_reset=0.50,
        )

        cp_map = set()
        edge_list = None
        if isinstance(inter_graph, list):
            edge_list = []
            for ctrl, trgt, _ in inter_graph:
                edge_list.append((ctrl, trgt))
        else:
            edge_list = inter_graph.edge_list()

        for c_instr in qc._data:
            qubits = c_instr.qubits
            if len(qubits) == 2:
                cp_map.update({(qubits[0]._index, qubits[1]._index)})

        # make sure every qubit-pair from the circuit actually present in the edge_list
        for cp in cp_map:
            self.assertIn(cp, edge_list)

    def test_2q_gates_excluded_edges_with_zero_weight(self):
        """Test 2Q gates are not applied to the qubit-pairs given by the interaction graph
        whose weight is zero"""

        num_qubits = 7
        pydi_graph = rx.PyDiGraph()
        pydi_graph.add_nodes_from(range(num_qubits))
        cp_map = [(0, 1, 10), (1, 2, 11), (2, 3, 0), (3, 4, 9), (4, 5, 12), (5, 6, 13)]
        pydi_graph.add_edges_from(cp_map)

        qc = random_circuit_from_graph(
            interaction_graph=pydi_graph,
        )
        ckt_cp_mp = set()
        for c_instr in qc._data:
            qubits = c_instr.qubits
            if len(qubits) == 2:
                ckt_cp_mp.update({(qubits[0]._index, qubits[1]._index)})

        # make sure qubit-pair with zero weight is not present in the edge_list from
        # the circuit.
        self.assertFalse((2, 3) in ckt_cp_mp)

    def test_edges_weight_with_some_None_raises(self):
        """Test if the function raises ValueError, if some of the edge
        weights are None, but not all."""

        pydi_graph = rx.PyDiGraph()
        pydi_graph.add_nodes_from(range(5))
        cp_mp = [(0, 1, None), (1, 2, 54), (2, 3, 23), (3, 4, 32)]

        pydi_graph.add_edges_from(cp_mp)
        with self.assertRaisesRegex(ValueError, ".getting selected is."):
            _ = random_circuit_from_graph(
                interaction_graph=pydi_graph,
            )

    def test_max_operands_not_between_1_2_raises(self):
        """Test if the function raises CircuitError when max_operands is not 1 or 2"""

        pydi_graph = rx.PyDiGraph()
        pydi_graph.add_nodes_from(range(10))
        with self.assertRaisesRegex(CircuitError, ".should be either."):
            _ = random_circuit_from_graph(
                interaction_graph=pydi_graph,
                min_2q_gate_per_edge=2,
                max_operands=3,  # This would fail
            )

    def test_negative_edge_weight_raises(self):
        """Test if negative edge weights raises ValueError"""

        pydi_graph = rx.PyDiGraph()
        pydi_graph.add_nodes_from(range(5))
        cp_mp = [(0, 1, -10), (1, 2, 54), (2, 3, 23), (3, 4, 32)]

        pydi_graph.add_edges_from(cp_mp)
        with self.assertRaisesRegex(ValueError, "Probabilities cannot be negative"):
            _ = random_circuit_from_graph(interaction_graph=pydi_graph)

    def test_raise_no_edges_insert_1q_oper_to_false(self):
        """Test if the function raises CircuitError when no edges are present in the
        interaction graph, which means there cannot be any 2Q gates, and only
        1Q gates present in the circuit, but `insert_1q_oper` is set to False"""
        inter_graph = rx.PyDiGraph()
        inter_graph.add_nodes_from(range(10))
        with self.assertRaisesRegex(CircuitError, ".there could be only 1Q gates."):
            _ = random_circuit_from_graph(
                interaction_graph=inter_graph,
                min_2q_gate_per_edge=2,
                conditional=True,
                reset=True,
                seed=0,
                insert_1q_oper=False,  # This will error out!
                prob_conditional=0.9,
                prob_reset=0.9,
            )

    def test_no_1q_when_insert_1q_oper_is_false(self):
        """Test no 1Q gates in the circuit, if `insert_1q_oper` is set to False."""
        num_qubits = 7
        pydi_graph = rx.PyDiGraph()
        pydi_graph.add_nodes_from(range(num_qubits))
        cp_map = [(0, 1, 10), (1, 2, 11), (2, 3, 0), (3, 4, 9), (4, 5, 12), (5, 6, 13)]
        pydi_graph.add_edges_from(cp_map)

        qc = random_circuit_from_graph(
            interaction_graph=pydi_graph,
            min_2q_gate_per_edge=2,
            measure=True,
            conditional=True,
            reset=True,
            seed=0,
            insert_1q_oper=False,
            prob_conditional=0.8,
            prob_reset=0.6,
        )

        count_1q = 0
        count_2q = 0

        for instr in qc:
            if instr.operation.name in {"measure", "delay", "reset"}:
                continue

            if instr.operation.num_qubits == 1:
                count_1q += 1
            if instr.operation.num_qubits == 2:
                count_2q += 1

        # 1Q gates should not be there in the circuits.
        self.assertEqual(count_1q, 0)

        # 2Q gates should be in the circuits.
        self.assertNotEqual(count_2q, 0)

    def test_conditionals_on_1q_operation(self):
        """Test if conditionals are present on 1Q operations"""

        num_qubits = 7
        pydi_graph = rx.PyDiGraph()
        pydi_graph.add_nodes_from(range(num_qubits))
        cp_map = [(0, 1, 10), (1, 2, 11), (2, 3, 0), (3, 4, 9), (4, 5, 12), (5, 6, 13)]
        pydi_graph.add_edges_from(cp_map)

        qc = random_circuit_from_graph(
            interaction_graph=pydi_graph,
            min_2q_gate_per_edge=4,
            measure=True,
            conditional=True,
            reset=True,
            seed=0,
            prob_conditional=0.91,
            prob_reset=0.6,
        )

        cond_counter_1q = 0
        cond_counter_2q = 0

        for instr in qc:
            if instr.operation.name in {"measure", "delay", "reset"}:
                continue

            cond = getattr(instr.operation, "_condition", None)
            if not cond is None:
                if instr.operation.num_qubits == 1:
                    cond_counter_1q += 1

                if instr.operation.num_qubits == 2:
                    cond_counter_2q += 1

        # Check if conditionals are present on 1Q and 2Q gates.
        self.assertNotEqual(cond_counter_1q, 0)
        self.assertNotEqual(cond_counter_2q, 0)

    def test_edges_prob(self):
        """Test if the probabilities of edges selected from the coupling
        map is indeed equal to the probabilities supplied with the coupling
        map, also test if for a sufficiently large circuit all edges in the
        coupling map is present in the circuit.
        """

        num_qubits = 3
        seed = 121
        h_h_g = rx.generators.directed_heavy_hex_graph(d=num_qubits, bidirectional=False)
        rng = np.random.default_rng(seed=seed)
        cp_map_list = []
        edge_list = h_h_g.edge_list()

        # generating a non-normalized list.
        list_choices = range(1, 100)  # keep the variance moderately high.
        random_probs = rng.choice(list_choices, size=len(edge_list)).tolist()
        sum_probs = sum(random_probs)

        for idx, qubits in enumerate(edge_list):
            ctrl, trgt = qubits
            cp_map_list.append((ctrl, trgt, random_probs[idx]))

        h_h_g.clear_edges()
        h_h_g.add_edges_from(cp_map_list)

        # The choices of probabilities are such that an edge might have a very low
        # probability of getting selected, so we have to generate a fairly big
        # circuit to include that edge in the circuit, and achieve the required
        # probability.
        qc = random_circuit_from_graph(
            h_h_g,
            min_2q_gate_per_edge=10,
            conditional=True,  # Just making it a bit more challenging.
            reset=True,
            seed=seed,
            insert_1q_oper=False,
            prob_conditional=0.91,
            prob_reset=0.50,
        )
        edge_count = defaultdict(int)

        for c_instr in qc._data:
            qubits = c_instr.qubits
            if len(qubits) == 2:
                edge_count[(qubits[0]._index, qubits[1]._index)] += 1

        # make sure every qubit-pair from the edge_list is present in the circuit.
        for ctrl, trgt, _ in cp_map_list:
            self.assertIn((ctrl, trgt), edge_count)

        edges_norm_qc = {}
        for edge, prob in edge_count.items():
            edges_norm_qc[edge] = prob / sum(edge_count.values())

        edges_norm_orig = {}
        for ctrl, trgt, prob in cp_map_list:
            edges_norm_orig[(ctrl, trgt)] = prob / sum_probs

        # Check if the probabilities of occurrences of edges in the circuit,
        # is indeed equal to the probabilities supplied as the edge data in
        # the interaction graph, upto a given tolerance.
        prob_deviations = []
        for edge_orig, prob_orig in edges_norm_orig.items():
            prob_deviations.append(np.absolute(prob_orig - edges_norm_qc[edge_orig]))

        # Setting 1% tolerance in probabilities.
        self.assertLess(max(prob_deviations), 0.01)

    def test_invalid_interaction_graph(self):
        """Test if CircuitError is raised when passed with an invalid interaction graph"""
        cp_map = [(0, 1, 10), (1, 2, 11), (2, 3, 0), (3, 4, 9), (4, 5, 12), (5, 6, 13)]
        with self.assertRaisesRegex(CircuitError, ".interaction graph object."):
            _ = random_circuit_from_graph(tuple(cp_map))  # Tuples are invalid interaction graph



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test cases for circuit annotations."""

# pylint: disable=missing-class-docstring,missing-function-docstring

from qiskit.circuit import annotation
from test import QiskitTestCase  # pylint: disable=wrong-import-order

# Most of the tests for this module are elsewhere, such as for OpenQASM 3 serialisation being in the
# OpenQASM 3 tests, and similar for QPY.


class TestAnnotation(QiskitTestCase):
    def test_iter_namespaces(self):
        self.assertEqual(
            list(annotation.iter_namespaces("hello.world")), ["hello.world", "hello", ""]
        )
        self.assertEqual(list(annotation.iter_namespaces("single_word")), ["single_word", ""])
        self.assertEqual(list(annotation.iter_namespaces("")), [""])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Synthesize a single qubit gate to a discrete basis set."""

from __future__ import annotations

import typing
import warnings
import numpy as np
from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit.gate import Gate
from qiskit.circuit.library import get_standard_gate_name_mapping, IGate
from qiskit.utils.deprecation import deprecate_func
from qiskit._accelerate.synthesis.discrete_basis import (
    SolovayKitaevSynthesis as RustSolovayKitaevSynthesis,
    GateSequence,
)

if typing.TYPE_CHECKING:
    from qiskit.dagcircuit import DAGCircuit


class SolovayKitaevDecomposition:
    """The Solovay Kitaev discrete decomposition algorithm.

    This class is called recursively by the transpiler pass, which is why it is separated.
    See :class:`~qiskit.transpiler.passes.SolovayKitaev` for more information.
    """

    def __init__(
        self,
        basic_approximations: str | dict[str, np.ndarray] | list[GateSequence] | None = None,
        *,
        basis_gates: list[str | Gate] | None = None,
        depth: int = 12,
        check_input: bool = False,
    ) -> None:
        """
        Args:
            basic_approximations: A specification of the basic SO(3) approximations in terms
                of discrete gates. At each iteration this algorithm, the remaining error is
                approximated with the closest sequence of gates in this set.
                If a ``str``, this specifies a filename from which to load the
                approximation. If a ``dict``, then this contains
                ``{gates: effective_SO3_matrix}`` pairs,
                e.g. ``{"h t": np.array([[0, 0.7071, -0.7071], [0, -0.7071, -0.7071], [-1, 0, 0]]}``.
                If a list, this contains the same information as the dict, but already converted to
                :class:`.GateSequence` objects, which contain the SO(3) matrix and gates.

                Either this parameter, or ``basis_gates`` and ``depth`` can be specified.
            basis_gates: A list of discrete (i.e., non-parameterized) standard gates.
                Defaults to ``["h", "t", "tdg"]``.
            depth: The number of basis gate combinations to consider in the basis set. This
                determines how fast (and if) the algorithm converges and should be chosen
                sufficiently high.
            check_input: If ``True``, perform intermediate steps checking whether the matrices
                are of expected form.
        """
        if basic_approximations is None:
            if basis_gates is not None:
                basis_gates = normalize_gates(basis_gates)
            self._sk = RustSolovayKitaevSynthesis(basis_gates, depth, None, check_input)

        elif basis_gates is not None:
            raise ValueError(
                "Either basic_approximations or basis_gates + depth can be specified, not both."
            )

        else:
            # Fast Rust path to load the file
            if isinstance(basic_approximations, str) and basic_approximations[~3:] != ".npy":
                self._sk = RustSolovayKitaevSynthesis.from_basic_approximations(
                    basic_approximations, True
                )
            else:
                sequences = self.load_basic_approximations(basic_approximations)
                self._sk = RustSolovayKitaevSynthesis.from_sequences(sequences, True)

        self._depth = depth
        self._check_input = check_input
        self._basis_gates = basis_gates

    @property
    def depth(self) -> int:
        """The maximum gate depth of the basic approximations."""
        return self._depth

    @property
    def check_input(self) -> bool:
        """Whether to perform runtime checks on the internal data."""
        return self._check_input

    @property
    def basis_gates(self) -> list[str] | None:
        """The basis gate set of the basic approximations.

        If ``None``, defaults to ``["h", "t", "tdg"]``.
        """
        return self._basis_gates

    @staticmethod
    def load_basic_approximations(data: list | str | dict) -> list[GateSequence]:
        """Load basic approximations.

        Args:
            data: If a string, specifies the path to the file from where to load the data.
                If a dictionary, directly specifies the decompositions as ``{gates: matrix}``
                or ``{gates: (matrix, global_phase)}``. There, ``gates`` are the names of the gates
                producing the SO(3) matrix ``matrix``, e.g.
                ``{"h t": np.array([[0, 0.7071, -0.7071], [0, -0.7071, -0.7071], [-1, 0, 0]]}``
                and the ``global_phase`` can be given to account for a global phase difference
                between the U(2) matrix of the quantum gates and the stored SO(3) matrix.
                If not given, the ``global_phase`` will be assumed to be 0.

        Returns:
            A list of basic approximations as type ``GateSequence``.

        Raises:
            ValueError: If the number of gate combinations and associated matrices does not match.
        """
        # new data format stored by the Rust internal class
        if isinstance(data, str) and data[-4:] != ".npy":
            sk = SolovayKitaevDecomposition(data)
            return sk._sk.get_gate_sequences()

        warnings.warn(
            "It is suggested to pass basic_approximations in the binary format produced "
            "by SolovayKitaevDecomposition.save_basic_approximations, which is more "
            "performant than other formats. Other formats are pending deprecation "
            "and will be deprecated in a future release.",
            category=PendingDeprecationWarning,
        )

        # is already a list of GateSequences
        if isinstance(data, list):
            return data

        # file is ``.npy``, load the dictionary it contains
        if isinstance(data, str):
            data = np.load(data, allow_pickle=True).item()

        # parse the dictionary
        sequences = []
        for gatestring, matrix_and_phase in data.items():
            if isinstance(matrix_and_phase, tuple):
                matrix, global_phase = matrix_and_phase
            else:
                matrix, global_phase = matrix_and_phase, 0

            # gates = [_1q_gates[element] for element in gatestring.split()]
            gates = normalize_gates(gatestring.split())
            sequence = GateSequence.from_gates_and_matrix(gates, matrix, global_phase)
            sequences.append(sequence)

        return sequences

    def save_basic_approximations(self, filename: str):
        """Save the basic approximations into a file.

        This can then be loaded again via the class initializer (preferred) or
        via :meth:`load_basic_approximations`::

            filename = "approximations.bin"
            sk.save_basic_approximations(filename)

            new_sk = SolovayKitaevDecomposition(filename)

        Args:
            filename: The filename to store the approximations in.

        Raises:
            ValueError: If the filename has a `.npy` extension. The format is not `.npy`,
                and storing as such can cause errors when loading the file again.
        """
        # Safety guard: previously, we serialized via npy, but this format is incompatible
        # with the current serialization, using Rust's serde + bincode. While we can still load
        # .npy files in legacy format, the new format should not be stored as .npy.
        if filename[~3:] == ".npy":
            raise ValueError(
                "The basic approximations are not stored in npy format. "
                "Choose a different file extension (e.g. .bin)."
            )
        self._sk.save_basic_approximations(filename)

    def run(
        self,
        gate_matrix: np.ndarray | Gate,
        recursion_degree: int,
        return_dag: bool = False,
        check_input: bool = True,
    ) -> QuantumCircuit | DAGCircuit:
        r"""Run the algorithm.

        Args:
            gate_matrix: The single-qubit gate to approximate. Can either be a :class:`.Gate`, where
                :meth:`.Gate.to_matrix` returns the matrix, or a :math:`2\times 2` unitary matrix
                representing the gate.
            recursion_degree: The recursion degree, called :math:`n` in the paper.
            return_dag: If ``True`` return a :class:`.DAGCircuit`, else a :class:`.QuantumCircuit`.
            check_input: If ``True`` check that the input matrix is valid for the decomposition.
                Overrides the class attribute with the same name, but only for this function call.

        Returns:
            A one-qubit circuit approximating the ``gate_matrix`` in the specified discrete basis.
        """
        # handle overriding the check_input setting
        self_check_input = self.check_input
        if check_input != self_check_input:
            self._sk.do_checks = check_input

        if isinstance(gate_matrix, Gate):
            data = self._sk.synthesize(gate_matrix, recursion_degree)
        else:
            data = self._sk.synthesize_matrix(gate_matrix, recursion_degree)

        if check_input != self_check_input:
            self._sk.do_checks = self_check_input

        circuit = QuantumCircuit._from_circuit_data(data, add_regs=True)

        if return_dag:
            from qiskit.converters import circuit_to_dag  # pylint: disable=cyclic-import

            return circuit_to_dag(circuit)

        return circuit

    def query_basic_approximation(self, gate: np.ndarray | Gate) -> QuantumCircuit:
        """Query a basic approximation of a matrix."""
        if isinstance(gate, Gate):
            data = self._sk.query_basic_approximation(gate)
        else:
            data = self._sk.query_basic_approximation_matrix(gate)

        circuit = QuantumCircuit._from_circuit_data(data, add_regs=True)
        return circuit

    @deprecate_func(
        since="2.1",
        additional_msg="Use query_basic_approximation instead, which takes a Gate or matrix "
        "as input and returns a QuantumCircuit object.",
        pending=True,
    )
    def find_basic_approximation(self, sequence: GateSequence) -> GateSequence:
        """Find ``GateSequence`` in ``self._basic_approximations`` that approximates ``sequence``.

        Args:
            sequence: ``GateSequence`` to find the approximation to.

        Returns:
            ``GateSequence`` in that approximates ``sequence``.
        """
        return self._sk.find_basic_approximation(sequence)


def normalize_gates(gates: list[Gate | str]) -> list[Gate]:
    """Normalize a list[Gate | str] into list[Gate]."""
    name_to_gate = get_standard_gate_name_mapping()
    # special case: we used to support "i" as IGate, but the official name is "id", so
    # we add it manually here
    name_to_gate["i"] = IGate()

    def normalize(gate: Gate | str) -> Gate:
        if isinstance(gate, Gate):
            return gate
        if gate in name_to_gate:
            return name_to_gate[gate]
        raise ValueError(f"Unsupported gate: {gate}")

    return list(map(normalize, gates))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Compute the sum of two qubit registers without any ancillary qubits."""

from __future__ import annotations
from math import ceil
from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit import QuantumRegister


def _mcx_ladder(n_mcx: int, alpha: int):
    r"""Implements a log-depth MCX ladder circuit as outlined in Algorithm 2 of [1]. The circuit
    relies on log-depth decomposition of MCX gate that uses conditionally clean ancillae of [2].
    Selecting :math:`\alpha=1` creates a CX ladder as shown in Fig. 2 of [1] and selecting
    :math:`\alpha=2` creates a Toffoli ladder as shown in Fig. 3 of [1].

    Args:
        n_mcx: Number of MCX gates in the ladder.
        alpha: Number of controls per MCX gate.

    Returns:
        QuantumCircuit: The MCX ladder circuit.

    References:
        1. Remaud and Vandaele, Ancilla-free Quantum Adder with Sublinear Depth, 2025.
        `arXiv:2501.16802 <https://arxiv.org/abs/2501.16802>`__

        2. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    def helper(qubit_indices, alphas) -> list[list[int]]:
        k = len(alphas) + 1
        if k == 1:
            return []
        if k == 2:
            return [qubit_indices[: alphas[0] + 1]]

        x_bar = [qubit_indices[alphas[0]]]
        alpha_bar = []
        right_pairs = [qubit_indices[0 : alphas[0] + 1]]
        left_pairs = [qubit_indices[alphas[k - 3] : alphas[-1] + 1]]

        for i in range(1, ceil(k / 2) - 1):
            left_pairs += [qubit_indices[alphas[2 * i - 2] : alphas[2 * i - 1] + 1]]
            right_pairs += [qubit_indices[alphas[2 * i - 1] : alphas[2 * i] + 1]]
            x_bar += (
                qubit_indices[alphas[2 * i - 2] + 1 : alphas[2 * i - 1]]
                + qubit_indices[alphas[2 * i - 1] + 1 : alphas[2 * i] + 1]
            )
            alpha_bar.append(alphas[2 * i] - alphas[0] - i)

        if k % 2 == 0:
            x_bar += qubit_indices[alphas[k - 4] + 1 : alphas[k - 3] + 1]
            alpha_bar.append(alphas[k - 3] - alphas[0] - int(k / 2) + 2)

        return left_pairs + helper(x_bar, alpha_bar) + right_pairs

    n = n_mcx * alpha + 1
    qc = QuantumCircuit(n)
    qubit_indices, alphas = list(range(n)), list(range(alpha, n, alpha))
    mcxs = helper(qubit_indices, alphas)
    for mcx in mcxs:
        qc.mcx(mcx[:-1], mcx[-1])

    return qc


def adder_ripple_r25(num_qubits: int) -> QuantumCircuit:
    # pylint: disable=line-too-long
    r"""The RV ripple carry adder [1].
    Construct an ancilla-free quantum adder circuit with sublinear depth based on the RV ripple-carry
    adder shown in [1]. The implementation has a depth of :math:`O(\log^2 n)` and uses
    math:`O(n \log n)` gates.

    As an example, a ripple-carry adder circuit that performs addition on two 4-qubit sized
    registers is as follows:

    .. parsed-literal::

                                       ┌───────────┐                    ┌────────┐
         a_0: ─────────────────────────┤0          ├────────────────────┤0       ├───────────────■─────────────────
                             ┌────────┐│           │                    │        │┌───────────┐  │
         a_1: ──■────────────┤0       ├┤2          ├──■─────────────────┤2       ├┤0          ├──┼────■────────────
                │            │        ││           │  │                 │        ││           │  │    │
         a_2: ──┼────■───────┤1       ├┤4          ├──┼────■────────────┤4       ├┤1 LAD_1_dg ├──┼────┼────■───────
                │    │       │        ││           │  │    │            │        ││           │  │    │    │
         a_3: ──┼────┼────■──┤2       ├┤6          ├──┼────┼────■───────┤6 LAD_2 ├┤2          ├──┼────┼────┼────■──
                │    │    │  │        ││           │  │    │    │       │        │└───────────┘┌─┴─┐  │    │    │
         b_0: ──┼────┼────┼──┤        ├┤1 LAD_2_dg ├──┼────┼────┼───────┤1       ├─────────────┤ X ├──┼────┼────┼──
              ┌─┴─┐  │    │  │  LAD_1 ││           │┌─┴─┐  │    │  ┌───┐│        │    ┌───┐    └───┘┌─┴─┐  │    │
         b_1: ┤ X ├──┼────┼──┤        ├┤3          ├┤ X ├──┼────┼──┤ X ├┤3       ├────┤ X ├─────────┤ X ├──┼────┼──
              └───┘┌─┴─┐  │  │        ││           │└───┘┌─┴─┐  │  ├───┤│        │    ├───┤         └───┘┌─┴─┐  │
         b_2: ─────┤ X ├──┼──┤        ├┤5          ├─────┤ X ├──┼──┤ X ├┤5       ├────┤ X ├──────────────┤ X ├──┼──
                   └───┘┌─┴─┐│        ││           │     └───┘┌─┴─┐└───┘└────────┘    └───┘              └───┘┌─┴─┐
         b_3: ──────────┤ X ├┤        ├┤7          ├──────────┤ X ├───────────────────────────────────────────┤ X ├
                        └───┘│        ││           │          └───┘                                           └───┘
        cout: ───────────────┤3       ├┤8          ├───────────────────────────────────────────────────────────────
                             └────────┘└───────────┘

    Here *LAD_1* and *LAD_2* are the CX and CCX ladders respectively introduced in [1]. Note that
    in this implementation the input register qubits are ordered as all qubits from
    the first input register, followed by all qubits from the second input register.

    Args:
        num_qubits: The size of the register.

    Returns:
        The quantum circuit implementing the RV ripple carry adder.

    Raises:
        ValueError: If ``num_state_qubits`` is lower than 1.

    **References:**

    1. Remaud and Vandaele, Ancilla-free Quantum Adder with Sublinear Depth, 2025.
    `arXiv:2501.16802 <https://arxiv.org/abs/2501.16802>`__

    """

    if num_qubits < 1:
        raise ValueError("The number of qubits must be at least 1.")

    qr_a = QuantumRegister(num_qubits, "a")
    qr_b = QuantumRegister(num_qubits, "b")
    qr_z = QuantumRegister(1, "cout")
    qc = QuantumCircuit(qr_a, qr_b, qr_z)

    if num_qubits == 1:
        qc.ccx(qr_a[0], qr_b[0], qr_z[0])
        qc.cx(qr_a[0], qr_b[0])
        return qc

    ab_interleaved = [q for pair in zip(qr_a, qr_b) for q in pair]

    qc.cx(qr_a[1:], qr_b[1:])
    qc.compose(_mcx_ladder(num_qubits - 1, 1), qr_a[1:] + qr_z[:], inplace=True)
    qc.compose(_mcx_ladder(num_qubits, 2).inverse(), ab_interleaved + qr_z[:], inplace=True)
    qc.cx(qr_a[1:], qr_b[1:])
    qc.x(qr_b[1 : max(2, num_qubits - 1)])
    qc.compose(_mcx_ladder(num_qubits - 1, 2), ab_interleaved[:-1], inplace=True)
    qc.x(qr_b[1 : max(2, num_qubits - 1)])
    qc.compose(_mcx_ladder(num_qubits - 2, 1).inverse(), qr_a[1:], inplace=True)
    qc.cx(qr_a, qr_b)

    return qc



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
======================================================
Circuit annotations (:mod:`qiskit.circuit.annotation`)
======================================================

.. currentmodule:: qiskit.circuit.annotation

This module contains the infrastructure for working with custom circuit annotations.

The main user-facing class is the base class :class:`qiskit.circuit.Annotation`, which is also
re-exported from this module.

.. _circuit-annotation-subclassing:

Custom annotation subclasses
============================

The :class:`.Annotation` class is intended to be subclassed.  Subclasses must set their
:attr:`~.Annotation.namespace` field.  This can be specific to an instance, or static for an entire
subclass.  The namespace is used as part of the dispatch mechanism, as described in
:ref:`circuit-annotation-namespacing`.

Circuit equality checks also compare annotations on objects in an order-dependent manner.  You will
likely want to implement the :meth:`~object.__eq__` magic method on any subclasses.

If you intend your annotation to be able to be serialized via :ref:`QPY <qiskit-qpy>` or :ref:`
OpenQASM 3 <qiskit-qasm3>`, you must provide separate implementations of the serialization and
deserialization methods as discussed in :ref:`circuit-annotation-serialization`.

.. _circuit-annotation-namespacing:

Namespacing
-----------

The "namespace" of an annotation is used as a look-up key when any consumer is deciding which
handler to invoke.  This includes in QPY and OpenQASM 3 serialization contexts, but in general,
transpiler passes will also look at annotations' namespaces to determine if they are relevant, and
so on.

This can be standard Python identifier (e.g. ``my_namespace``), or a dot-separated list of
identifiers (e.g. ``my_namespace.subnamespace``).  The namespace is used by all consumers of
annotations to determine what handler should be invoked.

A stand-alone function allows iterating through namespaces and parent namespaces in priority order
from most specific to least specific.

.. autofunction:: iter_namespaces


.. _circuit-annotation-serialization:

Serialization and deserialization
---------------------------------

Annotations represent completely custom data, that may persist after compilation.  This may include
data that should be serialized for later consumption, such as additional data that is interpreted by
a backend-compiler.  Qiskit's native binary QPY format (see :mod:`qiskit.qpy`) supports the concept
of arbitrary annotations in its payloads from version 15 onwards.  In OpenQASM 3 (see
:mod:`qiskit.qasm3`), annotations are a core language feature, and Qiskit's import/export support
for OpenQASM 3 includes serialization of annotations.

However, since annotations are generally custom subclasses and unknown to Qiskit, we cannot have
built-in support for serialization.  On the deserialization front, Qiskit will not, in general, have
an existing :class:`~.Annotation` object to call deserialization methods from.  It is also expected
that annotations may relate to some unknown-to-Qiskit shared state within a given circuit context.

For all of these reasons, serialization and deserialization of annotations is handled by custom
objects, which must be passed at the interface points of the relevant serialization functions.  For
example in QPY, the ``annotation_factories`` argument in :func:`.qpy.dump` and :func:`.qpy.load` are
used to pass serializers.

.. autoclass:: QPYSerializer
.. autoclass:: QPYFromOpenQASM3Serializer
.. autoclass:: OpenQASM3Serializer


Examples
========

A block-collection transpiler pass
----------------------------------

A principal goal of the annotation framework is to allow custom analyses and commands to be stored
on circuits in an instruction-local manner, either by the user on entry to the compiler, or for one
compiler pass to store information for later consumption.

For example, we can write a simple transpiler pass that collects runs of single-qubit operations,
and puts each run into a :class:`.BoxOp`, the calculates the total unitary action and attaches it as
a custom annotation, so the same analysis does not need to be repeated later, even if the internals
of each block are optimized.

.. code-block:: python

    from qiskit.circuit import annotation, QuantumCircuit, BoxOp
    from qiskit.quantum_info import Operator
    from qiskit.transpiler import TransformationPass

    class PerformsUnitary(annotation.Annotation):
        namespace = "unitary"
        def __init__(self, matrix):
            self.matrix = matrix

    class Collect1qRuns(TransformationPass):
        def run(self, dag):
            for run in dag.collect_1q_runs():
                block = QuantumCircuit(1)
                for node in run:
                    block.append(node.op, [0], [])
                box = BoxOp(block, annotations=[PerformsUnitary(Operator(block).data)])
                dag.replace_block_with_op(run, box, {run[0].qargs[0]: 0})
            return dag

In order to serialize the annotation to OpenQASM 3, we must define custom logic, since the analysis
itself is entirely custom.  The serialization is separate to the annotation; there may be
circumstances in which serialization should be done differently.

.. code-block:: python

    import ast
    import numpy as np

    class Serializer(annotation.OpenQASM3Serializer):
        def dump(self, annotation):
            if annotation.namespace != "unitary":
                return NotImplemented
            line = lambda row: "[" + ", ".join(repr(x) for x in row) + "]"
            return "[" + ", ".join(line(row) for row in annotation.matrix.tolist()) + "]"

        def load(self, namespace, payload):
            if namespace != "unitary":
                return NotImplemented
            return PerformsUnitary(np.array(ast.literal_eval(payload), dtype=complex))

Finally, this can be put together, showing the output OpenQASM 3.

.. code-block:: python

    from qiskit import qasm3

    qc = QuantumCircuit(3)
    qc.s(0)
    qc.t(0)
    qc.y(1)
    qc.x(1)
    qc.h(2)
    qc.s(2)
    collected = Collect1qRuns()(qc)

    handlers = {"unitary": Serializer()}
    dumped = qasm3.dumps(collected, annotation_handlers=handlers)
    print(dumped)

.. code-block:: openqasm3

    OPENQASM 3.0;
    include "stdgates.inc";
    qubit[3] q;
    @unitary [[(1+0j), 0j], [0j, (-0.7071067811865475+0.7071067811865475j)]]
    box {
      s q[0];
      t q[0];
    }
    @unitary [[1j, 0j], [0j, -1j]]
    box {
      y q[1];
      x q[1];
    }
    @unitary [[(0.7071067811865475+0j), (0.7071067811865475+0j)], [0.7071067811865475j, \
-0.7071067811865475j]]
    box {
      h q[2];
      s q[2];
    }

"""

from __future__ import annotations

import abc
from typing import Literal, Iterator

from qiskit._accelerate.circuit import Annotation


__all__ = [
    "Annotation",  # Also exported in `qiskit.circuit`, but for convenience is here too.
    "QPYSerializer",
    "OpenQASM3Serializer",
    "QPYFromOpenQASM3Serializer",
    "iter_namespaces",
]


def iter_namespaces(namespace: str) -> Iterator[str]:
    """An iterator over all namespaces that can be used to lookup the given namespace.

    This includes the namespace and all parents, including the root empty-string namespace.

    Examples:

    .. code-block:: python

        from qiskit.circuit.annotation import iter_namespaces
        assert list(iter_namespaces("hello.world")) == ["hello.world", "hello", ""]
    """
    while namespace:
        yield namespace
        split = namespace.rsplit(".", 1)
        if len(split) == 1:
            break
        namespace = split[0]
    yield ""


class QPYSerializer(abc.ABC):
    """The interface for serializers and deserializers of :class:`.Annotation` objects to QPY.

    For more information on QPY, see :mod:`qiskit.qpy`.

    This interface-definition class is designed to be subclassed.  The individual methods describe
    their contracts, and how they will be called.

    During QPY serialization and deserialization, the main QPY logic will call a factory function to
    create instances of subclasses of this class.  The return value from a given factory function
    will be used in *either* a serialization or deserialization context, but not both.

    The structure of calls during serialization of a single circuit is:

    1. many calls to :meth:`dump_annotation`, which will all share the same ``namespace`` argument,
       which will always be a (non-strict) prefix of all the :class:`.Annotation` objects given.
    2. one call to :meth:`dump_state`.

    The general structure of calls during deserialization of a single circuit out of a QPY payload
    is:

    1. one call to :meth:`load_state`, passing a ``namespace`` (with the same non-strict prefixing
       behavior as the "serializing" form).
    2. many calls to :meth:`load_annotation`, corresponding to annotations serialized under that
       namespace-prefix lookup.

    When subclassing this, recall that QPY is intended to have strict backwards-compatibility
    guarantees, and it is strongly recommended that annotation-serialisation subclasses maintain
    this.  In particular, it is suggested that any non-trivial serializer includes "version"
    information for the serializer in its total "state" (see :meth:`dump_state`), and the
    deserialization should make every effort to support backwards compatibility with previous
    versions of the same serializer.
    """

    @abc.abstractmethod
    def dump_annotation(
        self, namespace: str, annotation: Annotation
    ) -> bytes | Literal[NotImplemented]:
        """Serialize an annotation to a bytestream.

        This method may mutate the serializer's internal state (the object that will be serialized
        by :meth:`dump_state`).

        The ``namespace`` argument is the resolved key used to lookup this serializer.  It may not
        be identical to the :attr:`.Annotation.namespace` field of the ``annotation`` argument; it
        might be an ancestor, up to and including the empty string (the root namespace).  All calls
        to an instance of this class, as retrieved by a factory function in :func:`.qpy.dump` will
        be made using the same ``namespace``.

        The method can return :data:`NotImplemented` if the serializer cannot serialize a particular
        annotation.  In this case, the QPY logic will attempt to use a serializer registered for
        the parent namespaces.

        This method is the mirror of :meth:`load_annotation`.

        Args:
            namespace: the namespace that this serializer was accessed under.  This may be an
                ancestor of the annotation.
            annotation: the object to serialize to a bytestream.

        Returns:
            Either the serialized form of the annotation (optionally after mutating the
            serialization state of this class), or :data:`NotImplemented` if the annotation cannot
            be handled.
        """

    @abc.abstractmethod
    def load_annotation(self, payload: bytes) -> Annotation:
        """Load an annotation from a view of memory.

        A subclass can assume that :meth:`load_state` will have been called exactly once before
        this method is called, and all subsequent calls to this method will be for payloads
        corresponding to annotations serialized under that parent namespace.

        If a user configures QPY correctly, instances of this class will only be asked to
        deserialize payloads that the corresponding :meth:`dump_annotation` can successfully handle
        (i.e. return a payload, not :data:`NotImplemented`).  Subclasses may raise an arbitrary
        exception if this is not the case and this will abort the QPY load operation.  Such a
        situation would require that the user supplied a different serializer configuration on the
        two sides of the QPY load and dump.

        This method is the mirror of :meth:`dump_annotation`.

        Args:
            payload: the bytes to deserialized into an annotation.

        Returns:
            The deserialized annotation.
        """

    def dump_state(self) -> bytes:
        """Serialize a state object for the given serializer.

        When in a QPY dumping context, this method will be called exactly once, after all calls to
        :meth:`dump_annotation`.

        The default state is the empty bytestring; if your serializer is stateless, you do not need
        to override this method.

        This method is the mirror of :meth:`load_state`.
        """
        return b""

    def load_state(self, namespace: str, payload: bytes):  # pylint: disable=unused-argument
        """Initialize the state of the deserializer for a given ``namespace`` key.

        When in a QPY loading context, this method will be called exactly once, before all calls to
        :meth:`load_annotation`.  The ``namespace`` will be the same namespace that was passed to
        all calls to :meth:`dump_annotation` in the dumping context; that is, a (non-strict) prefix
        of the namespaces of all the :class:`.Annotation` objects its counterpart was asked to
        serialize.  For example, if the QPY dump was configured with::

            from qiskit import qpy
            from qiskit.circuit import annotation, Annotation

            class MyA(Annotation):
                namespace = "my.a"
            class MyB(Annotation):
                namespace = "my.b"

            class MyQPYSerializer(annotation.QPYSerializer):
                ...

            qpy.dump(..., annotation_factories={"my": MyQPYSerializer})

        then during the corresponding call to :func:`.qpy.load`, this method in ``MyQPYSerializer``
        will be called with ``"my"``, even though the annotations serialized had namespaces ``my.a``
        and ``my.b``. It is up to individual dumpers to do any sub-namespace handling they choose.

        The default implementation is a no-op; if you have not overridden :meth:`dump_state`, you do
        not need to override this method.

        This method is the mirror of :meth:`dump_state`.

        Args:
            namespace: the namespace key that the corresponding dump was resolved under.
            payload: the state payload that was dumped by the corresponding call to
                :meth:`dump_state`.
        """
        pass


class OpenQASM3Serializer(abc.ABC):
    """The interface for serializers and deserializers of :class:`.Annotation` objects to
    OpenQASM 3.

    For more information on OpenQASM 3 support in Qiskit, see :mod:`qiskit.qasm3`.

    This interface-definition class is designed to be subclassed.  OpenQASM 3 annotations are
    stateless within a program, therefore a subclass must not track state.
    """

    @abc.abstractmethod
    def dump(self, annotation: Annotation) -> str | Literal[NotImplemented]:
        """Serialize the paylaod of an annotation to a single line of UTF-8 text.

        The output of this method should not include the annotation's
        :attr:`~.Annotation.namespace` attribute; this is handled automatically by the OpenQASM 3
        exporter.

        The serialized form must not contain newline characters; it must be valid as the "arbitrary"
        component of the annotation as defined by OpenQASM 3.  If there is no data required, the
        method should return the empty string.  If this serializer cannot handle the particular
        annotation, it should return :data:`NotImplemented`.

        Args:
            annotation: the annotation object to serialize.

        Returns:
            the serialized annotation (without the namespace component), or the sentinel
            :data:`NotImplemented` if it cannot be handled by this object.
        """

    @abc.abstractmethod
    def load(self, namespace: str, payload: str) -> Annotation | Literal[NotImplemented]:
        """Load an annotation, if possible, from an OpenQASM 3 program.

        The two arguments will be the two components of an annotation, as defined by the OpenQASM 3
        specification.  The method should return :data:`NotImplemented` if it cannot handle the
        annotation.

        Args:
            namespace: the OpenQASM 3 "namespace" of the annotation.
            payload: the rest of the payload for the annotation.  This is arbitrary and free-form,
                and in general should have been serialized by a call to :meth:`dump`.

        Returns:
            the created :class:`.Annotation` object, whose :attr:`.Annotation.namespace` attribute
            should be identical to the incoming ``namespace`` argument.  If this class cannot handle
            the annotation, it can also return :data:`NotImplemented`.
        """

    def as_qpy(self) -> QPYFromOpenQASM3Serializer:
        """Derive a serializer/deserializer for QPY from this OpenQASM 3 variant.

        OpenQASM 3 serialization and deserialization is intended to be stateless and return single
        lines of UTF-8 encoded text.  This is a subset of the allowable serializations for QPY."""
        return QPYFromOpenQASM3Serializer(self)


class QPYFromOpenQASM3Serializer(QPYSerializer):
    """An adaptor that converts a :class:`OpenQASM3Serializer` into a :class:`QPYSerializer`.

    This works because OpenQASM 3 annotation serializers are required to be stateless and return
    UTF-8-encoded single lines of text, which is a subset of what QPY permits.

    Typically you create one of these using the :meth:`~OpenQASM3Serializer.as_qpy` method
    of an OpenQASM 3 annotation serializer.

    Examples:

    Instances of this class can be called like a zero-argument function and return themselves.  This
    lets you use them directly as a factory function to the QPY entry points, such as:

    .. code-block:: python

        import io
        from qiskit.circuit import OpenQASM3Serializer, Annotation
        from qiskit import qpy

        class MyAnnotation(Annotation):
            namespace = "my_namespace"

        class MySerializer(OpenQASM3Serializer):
            def dump(self, annotation):
                if not isinstance(annotation, MyAnnotation):
                    return NotImplemented
                return ""

            def load(self, namespace, payload):
                assert namespace == "my_namespace"
                assert payload == ""
                return MyAnnotation()

        qc = QuantumCircuit(2)
        with qc.box(annotations=[MyAnnotation()]):
            qc.cx(0, 1)

        with io.BytesIO() as fptr:
            qpy.dump(fptr, qc, annotation_serializers = {"my_namespace": MySerializer().as_qpy()})

    This is safe, without returning separate instances, because the base OpenQASM 3 serializers are
    necessarily stateless.
    """

    def __init__(self, inner: OpenQASM3Serializer):
        """
        Args:
            inner: the OpenQASM 3 serializer that this is derived from.
        """
        self.inner = inner

    def dump_annotation(self, namespace, annotation):
        qasm3 = self.inner.dump(annotation)
        if qasm3 is NotImplemented:
            return NotImplemented
        return (
            # For OpenQASM 3 serialisation, we need the exact namespace the annotation claims.
            annotation.namespace.encode("utf-8")
            + b"\x00"
            + qasm3.encode("utf-8")
        )

    def load_annotation(self, payload):
        namespace, payload = payload.split(b"\x00", maxsplit=1)
        out = self.inner.load(namespace.decode("utf-8"), payload.decode("utf-8"))
        if out is NotImplemented:
            raise ValueError(
                "asked to deserialize an object the provided OpenQASM deserializer cannot handle"
            )
        return out

    def __call__(self) -> QPYFromOpenQASM3Serializer:
        # Our internal object is stateless because it's an OpenQASM 3 exporter (which is a stateless
        # format).  Defining this method allows an instance of ourself to be used as a factory
        # function, simplifying the interface for creating a QPY serializer from an OpenQASM 3 one,
        # and attaching this to the class means that pickling would "just work".
        return self



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Rotation around an axis in x-y plane."""

import math
from cmath import exp
from typing import Optional
import numpy
from qiskit.circuit.gate import Gate
from qiskit.circuit.parameterexpression import ParameterValueType
from qiskit._accelerate.circuit import StandardGate


class RGate(Gate):
    r"""Rotation θ around the cos(φ)x + sin(φ)y axis.

    Can be applied to a :class:`~qiskit.circuit.QuantumCircuit`
    with the :meth:`~qiskit.circuit.QuantumCircuit.r` method.

    **Circuit symbol:**

    .. code-block:: text

               ┌─────────┐
        q_0:   ┤ R(θ,ϕ)  ├ 
               └─────────┘


    **Matrix Representation:**

    .. math::

        \newcommand{\rotationangle}{\frac{\theta}{2}}

        R(\theta, \phi) = e^{-i \rotationangle \left(\cos{\phi} x + \sin{\phi} y\right)} =
            \begin{pmatrix}
                \cos\left(\rotationangle\right) & -i e^{-i \phi} \sin\left(\rotationangle\right) \\
                -i e^{i \phi} \sin\left(\rotationangle\right) & \cos\left(\rotationangle\right)
            \end{pmatrix}
    """

    _standard_gate = StandardGate.R

    def __init__(
        self,
        theta: ParameterValueType,
        phi: ParameterValueType,
        label: Optional[str] = None,
    ):
        """Create new r single-qubit gate."""
        super().__init__("r", 1, [theta, phi], label=label)

    def _define(self):
        """Default definition"""
        # pylint: disable=cyclic-import
        from qiskit.circuit import QuantumCircuit

        #    ┌───────────────────────┐
        # q: ┤ U(θ,-π/2 + φ,π/2 - φ) ├
        #    └───────────────────────┘

        self.definition = QuantumCircuit._from_circuit_data(
            StandardGate.R._get_definition(self.params), add_regs=True, name=self.name
        )

    def inverse(self, annotated: bool = False):
        r"""Invert this gate as: :math:`R(θ, φ)^{\dagger} = R(-θ, φ)`

        Args:
            annotated: when set to ``True``, this is typically used to return an
                :class:`.AnnotatedOperation` with an inverse modifier set instead of a concrete
                :class:`.Gate`. However, for this class this argument is ignored as the inverse
                of this gate is always a :class:`.RGate` with an inverted parameter value.

        Returns:
            RGate: inverse gate.
        """
        return RGate(-self.params[0], self.params[1])

    def __array__(self, dtype=None, copy=None):
        """Return a numpy.array for the R gate."""
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        theta, phi = float(self.params[0]), float(self.params[1])
        cos = math.cos(theta / 2)
        sin = math.sin(theta / 2)
        exp_m = exp(-1j * phi)
        exp_p = exp(1j * phi)
        return numpy.array([[cos, -1j * exp_m * sin], [-1j * exp_p * sin, cos]], dtype=dtype)

    def power(self, exponent: float, annotated: bool = False):
        theta, phi = self.params
        return RGate(exponent * theta, phi)

    def __eq__(self, other):
        if isinstance(other, RGate):
            return self._compare_parameters(other)
        return False



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test synthesis algorithms for multi-controlled gates."""

import unittest
from test import combine
import numpy as np
from ddt import ddt, data

from qiskit.quantum_info import Operator
from qiskit.circuit import QuantumCircuit, Gate
from qiskit.circuit.library import (
    XGate,
    RXGate,
    RYGate,
    RZGate,
    PhaseGate,
    YGate,
    ZGate,
    HGate,
    SGate,
    SdgGate,
    TGate,
    TdgGate,
    SXGate,
    SXdgGate,
    UGate,
    U1Gate,
    U2Gate,
    U3Gate,
    CZGate,
)
from qiskit.synthesis.multi_controlled import (
    synth_mcx_n_dirty_i15,
    synth_mcx_n_clean_m15,
    synth_mcx_1_clean_b95,
    synth_mcx_1_clean_kg24,
    synth_mcx_1_dirty_kg24,
    synth_mcx_2_clean_kg24,
    synth_mcx_2_dirty_kg24,
    synth_mcx_gray_code,
    synth_mcx_noaux_v24,
    synth_c3x,
    synth_c4x,
)
from qiskit.circuit._utils import _compute_control_matrix
from qiskit.quantum_info.operators.operator_utils import _equal_with_ancillas, matrix_equal
from qiskit.transpiler import generate_preset_pass_manager

from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestMCSynthesisCorrectness(QiskitTestCase):
    """Test correctness of synthesis methods for multi-controlled gates."""

    @staticmethod
    def mc_matrix(base_gate: Gate, num_ctrl_qubits: int):
        """Return matrix for the MC gate with the given base gate and the number of control qubits."""
        base_mat = base_gate.to_matrix()
        return _compute_control_matrix(base_mat, num_ctrl_qubits)

    def assertSynthesisCorrect(
        self,
        base_gate: Gate,
        num_ctrl_qubits: int,
        synthesized_circuit: QuantumCircuit,
        clean_ancillas: bool,
    ):
        """Check correctness of a quantum circuit produced by a synthesis algorithm for multi-controlled
        gates, taking the additional ancilla qubits into account.

        This check is based on comparing the synthesized and the expected matrices and thus only
        works for synthesized circuits with up to about 10 qubits.

        Args:
            base_gate: the base gate of the MC gate.
            num_ctrl_qubits: the number of control qubits in the MC gate.
            synthesized_circuit: the quantum circuit synthesizing the MC gate.
            clean_ancillas: True if the algorithm uses clean ancilla qubits.

        Note: currently we do not have any MC synthesis algorithms that use both clean and dirty
        ancilla qubits. When we do, we will need to extend this function.
        """
        original_op = Operator(self.mc_matrix(base_gate, num_ctrl_qubits))
        synthesized_op = Operator(synthesized_circuit)

        num_qubits_original = original_op._op_shape._num_qargs_l
        num_qubits_synthesized = synthesized_circuit.num_qubits

        expected_op = Operator(
            np.kron(np.eye(2 ** (num_qubits_synthesized - num_qubits_original)), original_op)
        )
        if clean_ancillas:
            ancilla_qubits = list(range(num_qubits_original, num_qubits_synthesized))
        else:
            ancilla_qubits = []

        result = _equal_with_ancillas(
            synthesized_op,
            expected_op,
            ancilla_qubits,
        )
        self.assertTrue(result)

    @data(1, 2, 3, 4, 5, 6)
    def test_mcx_n_dirty_i15(self, num_ctrl_qubits: int):
        """Test synth_mcx_n_dirty_i15 by comparing synthesized and expected matrices."""
        synthesized_circuit = synth_mcx_n_dirty_i15(num_ctrl_qubits)
        self.assertSynthesisCorrect(
            XGate(), num_ctrl_qubits, synthesized_circuit, clean_ancillas=False
        )

    @data(3, 4, 5, 6)
    def test_mcx_n_clean_m15(self, num_ctrl_qubits: int):
        """Test synth_mcx_n_clean_m15 by comparing synthesized and expected matrices."""
        # Note: the method requires at least 3 control qubits
        synthesized_circuit = synth_mcx_n_clean_m15(num_ctrl_qubits)
        self.assertSynthesisCorrect(
            XGate(), num_ctrl_qubits, synthesized_circuit, clean_ancillas=True
        )

    @data(3, 4, 5, 6, 7, 8)
    def test_mcx_1_clean_b95(self, num_ctrl_qubits: int):
        """Test synth_mcx_1_clean_b95 by comparing synthesized and expected matrices."""
        # Note: the method requires at least 3 control qubits
        synthesized_circuit = synth_mcx_1_clean_b95(num_ctrl_qubits)
        self.assertSynthesisCorrect(
            XGate(), num_ctrl_qubits, synthesized_circuit, clean_ancillas=True
        )

    @data(3, 4, 5, 6, 7, 8)
    def test_mcx_1_clean_kg24(self, num_ctrl_qubits: int):
        """Test synth_mcx_1_clean_kg24 by comparing synthesized and expected matrices."""
        # Note: the method requires at least 3 control qubits
        synthesized_circuit = synth_mcx_1_clean_kg24(num_ctrl_qubits)
        self.assertSynthesisCorrect(
            XGate(), num_ctrl_qubits, synthesized_circuit, clean_ancillas=True
        )

    @data(3, 4, 5, 6, 7, 8)
    def test_mcx_1_dirty_kg24(self, num_ctrl_qubits: int):
        """Test synth_mcx_1_dirty_kg24 by comparing synthesized and expected matrices."""
        # Note: the method requires at least 3 control qubits
        synthesized_circuit = synth_mcx_1_dirty_kg24(num_ctrl_qubits)
        self.assertSynthesisCorrect(
            XGate(), num_ctrl_qubits, synthesized_circuit, clean_ancillas=False
        )

    @data(3, 4, 5, 6, 7, 8)
    def test_mcx_2_clean_kg24(self, num_ctrl_qubits: int):
        """Test synth_mcx_2_clean_kg24 by comparing synthesized and expected matrices."""
        # Note: the method requires at least 3 control qubits
        synthesized_circuit = synth_mcx_2_clean_kg24(num_ctrl_qubits)
        self.assertSynthesisCorrect(
            XGate(), num_ctrl_qubits, synthesized_circuit, clean_ancillas=True
        )

    @data(3, 4, 5, 6, 7, 8)
    def test_mcx_2_dirty_kg24(self, num_ctrl_qubits: int):
        """Test synth_mcx_2_dirty_kg24 by comparing synthesized and expected matrices."""
        # Note: the method requires at least 3 control qubits
        synthesized_circuit = synth_mcx_2_dirty_kg24(num_ctrl_qubits)
        self.assertSynthesisCorrect(
            XGate(), num_ctrl_qubits, synthesized_circuit, clean_ancillas=False
        )

    @data(3, 4, 5, 6, 7, 8)
    def test_mcx_gray_code(self, num_ctrl_qubits: int):
        """Test synth_mcx_gray_code by comparing synthesized and expected matrices."""
        # Note: the method requires at least 3 control qubits
        synthesized_circuit = synth_mcx_gray_code(num_ctrl_qubits)
        self.assertSynthesisCorrect(
            XGate(), num_ctrl_qubits, synthesized_circuit, clean_ancillas=False
        )

    @data(1, 2, 3, 4, 5, 6, 7, 8)
    def test_mcx_noaux_v24(self, num_ctrl_qubits: int):
        """Test synth_mcx_noaux_v24 by comparing synthesized and expected matrices."""
        synthesized_circuit = synth_mcx_noaux_v24(num_ctrl_qubits)
        self.assertSynthesisCorrect(
            XGate(), num_ctrl_qubits, synthesized_circuit, clean_ancillas=False
        )

    def test_c3x(self):
        """Test synth_c3x by comparing synthesized and expected matrices."""
        synthesized_circuit = synth_c3x()
        self.assertSynthesisCorrect(XGate(), 3, synthesized_circuit, clean_ancillas=False)

    def test_c4x(self):
        """Test synth_c4x by comparing synthesized and expected matrices."""
        synthesized_circuit = synth_c4x()
        self.assertSynthesisCorrect(XGate(), 4, synthesized_circuit, clean_ancillas=False)

    @combine(
        num_ctrl_qubits=[1, 2, 3, 4, 5, 6, 7],
        base_gate=[
            XGate(),
            YGate(),
            ZGate(),
            HGate(),
            PhaseGate(0.123),
            SGate(),
            SdgGate(),
            TGate(),
            TdgGate(),
            SXGate(),
            SXdgGate(),
            RXGate(0.789),
            RYGate(0.123),
            RZGate(0.456),
            UGate(0.1, 0.2, 0.3),
            U1Gate(0.1),
            U2Gate(0.1, 0.2),
            U3Gate(0.1, 0.2, 0.3),
        ],
        annotated=[False, True],
    )
    def test_create_mc_gates(self, num_ctrl_qubits, base_gate, annotated):
        """Test that creating various multi-controlled gates with small number of controls
        and no ancillas yields correct unitaries.
        """
        qc = QuantumCircuit(num_ctrl_qubits + 1)
        qc.append(
            base_gate.control(num_ctrl_qubits, annotated=annotated), range(num_ctrl_qubits + 1)
        )
        test_op = Operator(qc).data
        cop_mat = self.mc_matrix(base_gate, num_ctrl_qubits)
        self.assertTrue(matrix_equal(cop_mat, test_op))


@ddt
class TestMCSynthesisCounts(QiskitTestCase):
    """Test gate counts produced by multi-controlled synthesis methods."""

    def setUp(self):
        super().setUp()
        self.pm = generate_preset_pass_manager(
            optimization_level=0, basis_gates=["u", "cx"], seed_transpiler=12345
        )

    @data(5, 10, 15)
    def test_mcx_n_dirty_i15_cx_count(self, num_ctrl_qubits: int):
        """Test synth_mcx_n_dirty_i15 bound on CX count."""
        synthesized_circuit = synth_mcx_n_dirty_i15(num_ctrl_qubits)
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # The bound from the documentation of synth_mcx_n_dirty_i15
        self.assertLessEqual(cx_count, 8 * num_ctrl_qubits - 6)

    @data(5, 10, 15)
    def test_mcx_n_clean_m15_cx_count(self, num_ctrl_qubits: int):
        """Test synth_mcx_n_clean_m15 bound on CX count."""
        synthesized_circuit = synth_mcx_n_clean_m15(num_ctrl_qubits)
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # The bound from the documentation of synth_mcx_n_clean_m15
        self.assertLessEqual(cx_count, 6 * num_ctrl_qubits - 6)

    @data(5, 10, 15)
    def test_mcx_1_clean_b95_cx_count(self, num_ctrl_qubits: int):
        """Test synth_mcx_1_clean_b95 bound on CX count."""
        synthesized_circuit = synth_mcx_1_clean_b95(num_ctrl_qubits)
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # The bound from the documentation of synth_mcx_1_clean_b95
        self.assertLessEqual(cx_count, 16 * num_ctrl_qubits - 24)

    @data(3, 5, 10, 15)
    def test_mcx_1_clean_kg24_cx_count(self, num_ctrl_qubits: int):
        """Test synth_mcx_1_clean_kg24 bound on CX count."""
        synthesized_circuit = synth_mcx_1_clean_kg24(num_ctrl_qubits)
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # Based on the bound from the Sec 5.1 of arXiv:2407.17966, assuming relative-phase Toffoli
        # decomposition
        self.assertLessEqual(cx_count, 6 * num_ctrl_qubits - 6)

    @data(3, 5, 10, 15)
    def test_mcx_1_dirty_kg24_cx_count(self, num_ctrl_qubits: int):
        """Test synth_mcx_1_dirty_kg24 bound on CX count."""
        synthesized_circuit = synth_mcx_1_dirty_kg24(num_ctrl_qubits)
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        ## Based on the bound from the Sec 5.3 of arXiv:2407.17966, assuming relative-phase Toffoli
        # decomposition
        self.assertLessEqual(cx_count, 12 * num_ctrl_qubits - 18)

    @data(3, 5, 10, 15)
    def test_mcx_2_clean_kg24_cx_count(self, num_ctrl_qubits: int):
        """Test synth_mcx_2_clean_kg24 bound on CX count."""
        synthesized_circuit = synth_mcx_2_clean_kg24(num_ctrl_qubits)
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # Based on the bound from the Sec 5.2 of arXiv:2407.17966, assuming relative-phase Toffoli
        # decomposition
        self.assertLessEqual(cx_count, 6 * num_ctrl_qubits - 6)

    @data(3, 5, 10, 15)
    def test_mcx_2_dirty_kg24_cx_count(self, num_ctrl_qubits: int):
        """Test synth_mcx_2_dirty_kg24 bound on CX count."""
        synthesized_circuit = synth_mcx_2_dirty_kg24(num_ctrl_qubits)
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # Based on the bound from the Sec 5.4 of arXiv:2407.17966, assuming relative-phase Toffoli
        # decomposition
        self.assertLessEqual(cx_count, 12 * num_ctrl_qubits - 18)

    def test_c3x_cx_count(self):
        """Test synth_c3x bound on CX count."""
        synthesized_circuit = synth_c3x()
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # The bound from the default construction for C3X
        self.assertLessEqual(cx_count, 14)

    def test_c4x_cx_count(self):
        """Test synth_c4x bound on CX count."""
        synthesized_circuit = synth_c4x()
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # The bound from the default constuction for C4X
        self.assertLessEqual(cx_count, 36)

    @combine(
        num_ctrl_qubits=[5, 10, 15],
        base_gate=[RXGate(0.123), RYGate(0.456), RZGate(0.789)],
        annotated=[False, True],
    )
    def test_mc_rotation_gates_cx_count(
        self, num_ctrl_qubits: int, base_gate: Gate, annotated: bool
    ):
        """Test bounds on the number of CX gates for mcrx / mcry / mcrz."""
        qc = QuantumCircuit(num_ctrl_qubits + 1)
        qc.append(
            base_gate.control(num_ctrl_qubits, annotated=annotated), range(num_ctrl_qubits + 1)
        )
        transpiled_circuit = self.pm.run(qc)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # The synthesis of mcrx/mcry/mcrz gates uses _mcsu2_real_diagonal.
        # The bounds are given in arXiv:2302.06377, Theorem 3.
        # In practice, we actually get better bounds for small values of num_ctrl_qubits.
        expected_cx_count = 16 * (num_ctrl_qubits + 1) - 40
        self.assertLessEqual(cx_count, expected_cx_count)

    @data(5, 10, 15)
    def test_mcx_noaux_v24_cx_count(self, num_ctrl_qubits: int):
        """Test synth_mcx_noaux_v24 bound on CX count."""
        synthesized_circuit = synth_mcx_noaux_v24(num_ctrl_qubits)
        transpiled_circuit = self.pm.run(synthesized_circuit)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # The algorithm synth_mcx_noaux_v24 is based on the synthesis of MCPhase,
        # which is defined using a sequence of MCRZ gates:
        # MCPhase(n) is defined using one MCRZ(1), one MCRZ(2), ..., one MCRZ(n).
        # The bound below follows using the bound of 16*(k+1)-40 for MCRZ(k) and summing
        # the resulting arithmetic progression:
        #   sum_{k=1}^n (16*(k+1)-40) = sum_{k=1}^n (16*k - 24) =
        #     16*n*(n+1)/2 - 24*n = 8n^2 - 16*n.
        self.assertLessEqual(cx_count, 8 * num_ctrl_qubits**2 - 16 * num_ctrl_qubits)

    @combine(
        num_ctrl_qubits=[5, 8, 10, 13, 15],
        base_gate=[
            XGate(),
            YGate(),
            ZGate(),
            HGate(),
            SGate(),
            SdgGate(),
            TGate(),
            TdgGate(),
            SXGate(),
            SXdgGate(),
            PhaseGate(0.345),
        ],
        annotated=[False, True],
    )
    def test_mcx_equiv_noaux_cx_count(self, num_ctrl_qubits: int, base_gate: Gate, annotated: bool):
        """Test bounds on the number of CX-gates when synthesizing multi-controlled gates
        which are locally equivalent to MCX.
        """
        qc = QuantumCircuit(num_ctrl_qubits + 1)
        qc.append(
            base_gate.control(num_ctrl_qubits, annotated=annotated), range(num_ctrl_qubits + 1)
        )
        transpiled_circuit = self.pm.run(qc)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # The bounds should be the same as for synth_mcx_noaux_v24
        self.assertLessEqual(cx_count, 8 * num_ctrl_qubits**2 - 16 * num_ctrl_qubits)

    @combine(
        num_ctrl_qubits=[5, 10, 15],
        annotated=[False, True],
    )
    def test_mcu_noaux_cx_count(self, num_ctrl_qubits: int, annotated: bool):
        """Test bounds on the number of CX-gates when synthesizing multi-controlled single-qubit
        unitary gates.
        """
        base_gate = UGate(0.123, 0.456, 0.789)
        qc = QuantumCircuit(num_ctrl_qubits + 1)
        qc.append(
            base_gate.control(num_ctrl_qubits, annotated=annotated), range(num_ctrl_qubits + 1)
        )
        transpiled_circuit = self.pm.run(qc)
        cx_count = transpiled_circuit.count_ops()["cx"]
        # The synthesis of MCX(n) uses two MCRZ(n), one MCRY(n), and one MCPhase(n-1).
        # Thus the number of CX-gate should be upper-bounded by
        # 3*(16 * (n + 1) - 40) + (8 * (n-1)^2 - 16 * (n-1))
        self.assertLessEqual(cx_count, 8 * num_ctrl_qubits**2 + 16 * num_ctrl_qubits - 96)

    @combine(
        num_ctrl_qubits=[1, 2, 3, 4, 5, 6, 7, 8],
        base_gate=[
            XGate(),
            YGate(),
            ZGate(),
            HGate(),
            PhaseGate(0.123),
            SGate(),
            SdgGate(),
            TGate(),
            TdgGate(),
            SXGate(),
            SXdgGate(),
            RXGate(0.789),
            RYGate(0.123),
            RZGate(0.456),
            UGate(0.1, 0.2, 0.3),
            U1Gate(0.1),
            U2Gate(0.1, 0.2),
            U3Gate(0.1, 0.2, 0.3),
            CZGate(),
        ],
        annotated=[False, True],
    )
    def test_small_mc_gates_cx_count(self, num_ctrl_qubits: int, base_gate: Gate, annotated: bool):
        """Test that transpiling various multi-controlled gates with small number of controls and no
        ancillas yields the expected number of CX gates.

        This test prevents making changes to the synthesis algorithms that would deteriorate the
        quality of the synthesized circuits.
        """
        qc = QuantumCircuit(num_ctrl_qubits + base_gate.num_qubits)
        qc.append(base_gate.control(num_ctrl_qubits, annotated=annotated), qc.qubits)
        transpiled_circuit = self.pm.run(qc)
        cx_count = transpiled_circuit.count_ops()["cx"]

        if isinstance(base_gate, (XGate, YGate, ZGate, HGate)):
            # MCX gate and other locally equivalent multi-controlled gates
            expected = {1: 1, 2: 6, 3: 14, 4: 36, 5: 84, 6: 140, 7: 220, 8: 324}
        elif isinstance(
            base_gate, (PhaseGate, SGate, SdgGate, TGate, TdgGate, SXGate, SXdgGate, U1Gate)
        ):
            # MCPhase gate and other locally equivalent multi-controlled gates
            expected = {1: 2, 2: 6, 3: 20, 4: 44, 5: 84, 6: 140, 7: 220, 8: 324}
        elif isinstance(base_gate, RZGate):
            expected = {1: 2, 2: 4, 3: 14, 4: 24, 5: 40, 6: 56, 7: 80, 8: 104}
        elif isinstance(base_gate, (RXGate, RYGate)):
            expected = {1: 2, 2: 8, 3: 20, 4: 24, 5: 40, 6: 56, 7: 80, 8: 104}
        elif isinstance(base_gate, (UGate, U2Gate, U3Gate)):
            expected = {1: 2, 2: 22, 3: 54, 4: 92, 5: 164, 6: 252, 7: 380, 8: 532}
        elif isinstance(base_gate, CZGate):
            expected = {1: 6, 2: 14, 3: 36, 4: 84, 5: 140, 6: 220, 7: 324, 8: 444}
        else:
            raise NotImplementedError

        self.assertLessEqual(cx_count, expected[num_ctrl_qubits])


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Matplotlib circuit visualization style."""

from __future__ import annotations

import json
from pathlib import Path
from qiskit.visualization.circuit.qcstyle import StyleDict, DefaultStyle


class DAGStyleDict(StyleDict):
    """A dictionary for graphviz styles.

    Defines additional abbreviations for key accesses, such as allowing
    ``"ec"`` instead of writing ``"edgecolor"``.
    """

    VALID_FIELDS = {
        "name",
        "fontsize",
        "bgcolor",
        "dpi",
        "pad",
        "nodecolor",
        "inputnodecolor",
        "inputnodefontcolor",
        "outputnodecolor",
        "outputnodefontcolor",
        "opnodecolor",
        "opnodefontcolor",
        "edgecolor",
        "qubitedgecolor",
        "clbitedgecolor",
    }

    ABBREVIATIONS = {
        "nc": "nodecolor",
        "ic": "inputnodecolor",
        "if": "inputnodefontcolor",
        "oc": "outputnodecolor",
        "of": "outputnodefontcolor",
        "opc": "opnodecolor",
        "opf": "opnodefontcolor",
        "ec": "edgecolor",
        "qec": "qubitedgecolor",
        "clc": "clbitedgecolor",
    }


class DAGDefaultStyle(DefaultStyle):
    """Creates a Default Style dictionary

    The style dict contains numerous options that define the style of the
    output circuit visualization. The style dict is used by the `graphviz`
    output. The options available in the style dict are defined below:

    Attributes:
        name (str): The name of the style.
        fontsize (str): The font size to use for text.
        bgcolor (str): The color name to use for the background ('red', 'green', etc.).
        nodecolor (str): The color to use for all nodes.
        dpi (int): The DPI to use for the output image.
        pad (int): A number to adjust padding around output
            graph.
        inputnodecolor (str): The color to use for incoming wire nodes. Overrides
            nodecolor for those nodes.
        inputnodefontcolor (str): The font color to use for incoming wire nodes.
            Overrides nodecolor for those nodes.
        outputnodecolor (str): The color to use for output wire nodes. Overrides
            nodecolor for those nodes.
        outputnodefontcolor (str): The font color to use for output wire nodes.
            Overrides nodecolor for those nodes.
        opnodecolor (str): The color to use for Instruction nodes. Overrides
            nodecolor for those nodes.
        opnodefontcolor (str): The font color to use for Instruction nodes.
            Overrides nodecolor for those nodes.

        qubitedgecolor (str): The edge color for qubits. Overrides edgecolor for these edges.
        clbitedgecolor (str): The edge color for clbits. Overrides edgecolor for these edges.
    """

    DEFAULT_STYLE_NAME = "plain"
    DEFAULT_STYLE_PATH = Path(__file__).parent / "styles"

    def __init__(self):
        path = self.DEFAULT_STYLE_PATH / Path(self.DEFAULT_STYLE_NAME).with_suffix(".json")

        with open(path, "r") as infile:
            default_style = json.load(infile)

        # set shortcuts, such as "ec" for "edgecolor"
        self.style = DAGStyleDict(**default_style)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for cosine-sine decomposition."""

import math

from ddt import ddt
import numpy as np

from qiskit.quantum_info.random import random_unitary
from qiskit.quantum_info.operators.predicates import is_unitary_matrix
from qiskit.circuit._utils import _compute_control_matrix
from qiskit._accelerate.cos_sin_decomp import cossin

from test import combine  # pylint: disable=wrong-import-order
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestCSD(QiskitTestCase):
    """Test the cosine-sine decomposition."""

    def assertCossinDecompositionIsCorrect(self, mat):
        """Run QSD on mat and verify that it is correct."""
        n = mat.shape[0] // 2

        # Run QSD: u = [u1, u2], v = [v1, v2]
        csd_u, csd_thetas, csd_v = cossin(mat)

        # Check that the computed matrices are unitary
        self.assertTrue(is_unitary_matrix(csd_u[0], atol=1e-13))
        self.assertTrue(is_unitary_matrix(csd_u[1], atol=1e-13))
        self.assertTrue(is_unitary_matrix(csd_v[0], atol=1e-13))
        self.assertTrue(is_unitary_matrix(csd_v[1], atol=1e-13))

        # Create appropriate 2n x 2n matrices and multiply
        zero_mat = np.zeros((n, n), dtype=complex)
        c_mat = np.diag([math.cos(theta) for theta in csd_thetas])
        s_mat = np.diag([math.sin(theta) for theta in csd_thetas])
        cs_mat = np.vstack([np.hstack([c_mat, -1 * s_mat]), np.hstack([s_mat, c_mat])])
        u_mat = np.vstack([np.hstack([csd_u[0], zero_mat]), np.hstack([zero_mat, csd_u[1]])])
        v_mat = np.vstack([np.hstack([csd_v[0], zero_mat]), np.hstack([zero_mat, csd_v[1]])])
        recomputed_mat = u_mat @ cs_mat @ v_mat

        np.testing.assert_allclose(
            mat,
            recomputed_mat,
            atol=1e-7,
        )

    @combine(num_qubits=[1, 2, 3, 4], seed=list(range(100)))
    def test_random_unitary(self, num_qubits, seed):
        """Test CSD for random unitary matrices."""
        mat = random_unitary(2**num_qubits, seed=seed).data
        self.assertCossinDecompositionIsCorrect(mat)

    @combine(num_controls=[1, 2, 3], num_targets=[1, 2, 3], seed=list(range(50)))
    def test_controlled_random_unitary(self, num_controls, num_targets, seed):
        """Test CSD for controlled random unitary matrices."""
        # Note that scipy has numerical stability problems on controlled
        # random unitary matrices.
        base_mat = random_unitary(2**num_targets, seed=seed).data
        mat = _compute_control_matrix(base_mat, num_controls)
        self.assertCossinDecompositionIsCorrect(mat)

    @combine(num_qubits=[1, 2, 3, 4], seed=list(range(20)))
    def test_random_hermitian(self, num_qubits, seed):
        """Test CSD for random Hermitian matrices."""
        umat = random_unitary(2**num_qubits, seed=seed).data
        np.random.seed(seed)
        dmat = np.diag(np.exp(1j * np.random.normal(size=2**num_qubits)))
        mat = umat.T.conjugate() @ dmat @ umat
        self.assertCossinDecompositionIsCorrect(mat)

    @combine(num_qubits=[1, 2, 3, 4], seed=list(range(20)))
    def test_random_diagonal(self, num_qubits, seed):
        """Test CSD for random diagonal matrices."""
        np.random.seed(seed)
        dmat = np.diag(np.exp(1j * np.random.normal(size=2**num_qubits)))
        self.assertCossinDecompositionIsCorrect(dmat)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Matplotlib circuit visualization style."""

from __future__ import annotations

import json
from pathlib import Path

from qiskit.visualization.style import StyleDict, DefaultStyle


class MPLStyleDict(StyleDict):
    """A dictionary for matplotlib styles.

    Defines additional abbreviations for key accesses, such as allowing
    ``"ec"`` instead of writing ``"edgecolor"``.
    """

    VALID_FIELDS = {
        "name",
        "textcolor",
        "gatetextcolor",
        "subtextcolor",
        "linecolor",
        "creglinecolor",
        "gatefacecolor",
        "barrierfacecolor",
        "backgroundcolor",
        "edgecolor",
        "fontsize",
        "subfontsize",
        "showindex",
        "figwidth",
        "dpi",
        "margin",
        "creglinestyle",
        "displaytext",
        "displaycolor",
    }

    ABBREVIATIONS = {
        "tc": "textcolor",
        "gt": "gatetextcolor",
        "sc": "subtextcolor",
        "lc": "linecolor",
        "cc": "creglinecolor",
        "gc": "gatefacecolor",
        "bc": "barrierfacecolor",
        "bg": "backgroundcolor",
        "ec": "edgecolor",
        "fs": "fontsize",
        "sfs": "subfontsize",
        "index": "showindex",
        "cline": "creglinestyle",
        "disptex": "displaytext",
        "dispcol": "displaycolor",
    }

    NESTED_ATTRS = {"displaycolor", "displaytext"}


class MPLDefaultStyle(DefaultStyle):
    """Creates a Default Style dictionary

    The style dict contains numerous options that define the style of the
    output circuit visualization. The style dict is used by the `mpl` or
    `latex` output. The options available in the style dict are defined below:

    Attributes:
        name (str): The name of the style. The name can be set to ``iqp``,
            ``iqp-dark``, ``textbook``, ``bw``, ``clifford``, or the name of a
            user-created json file. This overrides the setting in the user config
            file (usually ``~/.qiskit/settings.conf``).
        textcolor (str): the color code to use for all text not inside a gate.
        subtextcolor (str): the color code to use for subtext.
        linecolor (str): the color code to use for lines.
        creglinecolor (str): The color code to use for classical register lines.
        gatetextcolor (str): The color code to use for gate text.
        gatefacecolor (str): The color code to use for a gate if no color
            specified in the 'displaycolor' dict.
        barrierfacecolor (str): The color code to use for barriers.
        backgroundcolor (str): The color code to use for the background.
        edgecolor (str): The color code to use for gate edges when using the
            `bw` style.
        fontsize (int): The font size to use for text.
        subfontsize (int): The font size to use for subtext.
        showindex (bool): If set to True, show the index numbers at the top.
        figwidth (int): The maximum width (in inches) for the output figure.
            If set to -1, the maximum displayable width will be used.
        dpi (int): The DPI to use for the output image.
        margin (list): A list of margin values to adjust spacing around output
            image. Takes a list of 4 ints: [x left, x right, y bottom, y top].
        creglinestyle (str): The style of line to use for classical registers.
            Choices are ``solid``, ``doublet``, or any valid matplotlib
            `linestyle` kwarg value.
        displaytext (dict): a dictionary of the text to use for certain element
            types in the output visualization. These items allow the use of
            LaTeX formatting for gate names. The 'displaytext' dict can contain
            any number of elements. User created names and labels may be used as
            keys, which allow these to have Latex formatting.
        displaycolor (dict): the color codes to use for each circuit element in
            the form (gate_color, text_color). Colors can also be entered without
            the text color, such as 'u1': '#FA74A6', in which case the text color
            will always be `gatetextcolor`. The `displaycolor` dict can contain
            any number of elements. User names and labels may be used as keys,
            which allows for custom colors for user-created gates.
    """

    DEFAULT_STYLE_NAME = "iqp"
    DEFAULT_STYLE_PATH = Path(__file__).parent / "styles"

    def __init__(self):
        path = self.DEFAULT_STYLE_PATH / Path(self.DEFAULT_STYLE_NAME).with_suffix(".json")

        with open(path, "r") as infile:
            default_style = json.load(infile)

        # set shortcuts, such as "ec" for "edgecolor"
        self.style = MPLStyleDict(**default_style)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the Solovay Kitaev transpilation pass."""

import os
import unittest
import tempfile
import numpy as np
import scipy

from ddt import ddt, data

from qiskit import transpile
from qiskit.circuit import QuantumCircuit, Parameter
from qiskit.circuit import ClassicalRegister
from qiskit.circuit.library import (
    TGate,
    TdgGate,
    HGate,
    SGate,
    SdgGate,
    QFTGate,
    RXGate,
    RYGate,
    RZGate,
)
from qiskit.circuit import QuantumRegister
from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.quantum_info import Operator
from qiskit.synthesis.discrete_basis.generate_basis_approximations import (
    generate_basic_approximations,
)
from qiskit.transpiler import PassManager
from qiskit.transpiler.passes import UnitarySynthesis, Collect1qRuns, ConsolidateBlocks
from qiskit.transpiler.passes.synthesis import SolovayKitaev, SolovayKitaevSynthesis
from qiskit.synthesis.discrete_basis import SolovayKitaevDecomposition
from test import QiskitTestCase, combine  # pylint: disable=wrong-import-order


def _trace_distance(circuit1, circuit2):
    """Return the trace distance of the two input circuits."""
    op1, op2 = Operator(circuit1), Operator(circuit2)
    return 0.5 * np.trace(scipy.linalg.sqrtm(np.conj(op1 - op2).T.dot(op1 - op2))).real


def is_so3_matrix(array: np.ndarray) -> bool:
    """Check if the input array is a SO(3) matrix."""
    if array.shape != (3, 3):
        return False

    if abs(np.linalg.det(array) - 1.0) > 1e-10:
        return False

    if False in np.isreal(array):
        return False

    return True


@ddt
class TestSolovayKitaev(QiskitTestCase):
    """Test the Solovay Kitaev algorithm and transformation pass."""

    def setUp(self):
        super().setUp()

        self.basic_approx = generate_basic_approximations([HGate(), TGate(), TdgGate()], 3)
        self.default_sk = SolovayKitaev()

    def test_unitary_synthesis(self):
        """Test the unitary synthesis transpiler pass with Solovay-Kitaev."""
        circuit = QuantumCircuit(2)
        circuit.rx(0.8, 0)
        circuit.cx(0, 1)
        circuit.x(1)

        _1q = Collect1qRuns()
        _cons = ConsolidateBlocks()
        _synth = UnitarySynthesis(["h", "t", "tdg"], method="sk")
        passes = PassManager([_1q, _cons, _synth])
        compiled = passes.run(circuit)

        diff = _trace_distance(circuit, compiled)
        self.assertLess(diff, 1e-4)
        self.assertEqual(set(compiled.count_ops().keys()), {"h", "t", "tdg", "cx"})

    def test_plugin(self):
        """Test calling the plugin directly."""
        circuit = QuantumCircuit(1)
        circuit.rx(0.8, 0)

        unitary = Operator(circuit).data

        plugin = SolovayKitaevSynthesis()
        out = plugin.run(unitary, basis_gates=["h", "s"], depth=10, recursion_degree=3)

        self.assertTrue(set(out.count_ops().keys()).issubset({"h", "s", "sdg", "cx"}))

    def test_multiple_plugins(self):
        """Test calling the plugins directly but with different instances of basis set."""
        circuit = QuantumCircuit(1)
        circuit.rx(0.8, 0)
        unitary = Operator(circuit).data
        plugin = SolovayKitaevSynthesis()

        # First, use the ["h", "t", "tdg"] set
        out = plugin.run(unitary, basis_gates=["h", "t", "tdg"])
        self.assertTrue(set(out.count_ops().keys()).issubset({"h", "t", "tdg"}))

        # Second, use the ["h", "s"] set
        out = plugin.run(unitary, basis_gates=["h", "s", "sdg"])
        self.assertTrue(set(out.count_ops().keys()).issubset({"h", "s", "sdg"}))

        # Third, use the ["h", "t", "tdg"] set again
        out = plugin.run(unitary, basis_gates=["h", "t", "tdg"])
        self.assertTrue(set(out.count_ops().keys()).issubset({"h", "t", "tdg"}))

    def test_i_returns_empty_circuit(self):
        """Test that ``SolovayKitaev`` returns an empty circuit when
        it approximates the I-gate."""
        circuit = QuantumCircuit(1)
        circuit.id(0)

        decomposed_circuit = self.default_sk(circuit)
        self.assertEqual(QuantumCircuit(1), decomposed_circuit)

    def test_exact_decomposition_acts_trivially(self):
        """Test that the a circuit that can be represented exactly is represented exactly."""
        circuit = QuantumCircuit(1)
        circuit.t(0)
        circuit.h(0)
        circuit.tdg(0)

        sk = SolovayKitaev(3, basis_gates=["h", "t", "tdg"], depth=12)

        dag = circuit_to_dag(circuit)
        decomposed_dag = sk.run(dag)
        decomposed_circuit = dag_to_circuit(decomposed_dag)
        self.assertEqual(circuit, decomposed_circuit)

    def test_str_basis_gates(self):
        """Test specifying the basis gates by string works."""
        circuit = QuantumCircuit(1)
        circuit.rx(0.8, 0)

        basic_approx = generate_basic_approximations(["h", "t", "s"], 3)
        sk = SolovayKitaev(3, basic_approx)

        dag = circuit_to_dag(circuit)
        discretized = dag_to_circuit(sk.run(dag))

        reference = QuantumCircuit(1, global_phase=15 * np.pi / 8)
        reference.h(0)
        reference.t(0)
        reference.h(0)

        # Make sure that the discretized circuit gives a valid approximation
        diff = _trace_distance(circuit, discretized)
        self.assertLess(diff, 0.01)
        self.assertEqual(discretized, reference)

    def test_approximation_on_qft(self):
        """Test the Solovay-Kitaev decomposition on the QFT circuit."""
        qft = QuantumCircuit(3)
        qft.append(QFTGate(3), [0, 1, 2], [])
        transpiled = transpile(
            qft, basis_gates=["u1", "u2", "u3", "u", "cx", "id"], optimization_level=1
        )

        skd = SolovayKitaev(5)

        with self.subTest("1 recursion"):
            discretized = skd(transpiled)
            self.assertLess(_trace_distance(transpiled, discretized), 15)

        skd.recursion_degree = 5
        with self.subTest("2 recursions"):
            discretized = skd(transpiled)
            self.assertLess(_trace_distance(transpiled, discretized), 7)

    def test_u_gates_work(self):
        """Test SK works on Qiskit's UGate.

        Regression test of Qiskit/qiskit-terra#9437.
        """
        circuit = QuantumCircuit(1)
        circuit.u(np.pi / 2, -np.pi, -np.pi, 0)
        circuit.u(np.pi / 2, np.pi / 2, -np.pi, 0)
        circuit.u(-np.pi / 4, 0, -np.pi / 2, 0)
        circuit.u(np.pi / 4, -np.pi / 16, 0, 0)
        circuit.u(0, 0, np.pi / 16, 0)
        circuit.u(0, np.pi / 4, np.pi / 4, 0)
        circuit.u(np.pi / 2, 0, -15 * np.pi / 16, 0)
        circuit.p(-np.pi / 4, 0)
        circuit.p(np.pi / 4, 0)
        circuit.u(np.pi / 2, 0, -3 * np.pi / 4, 0)
        circuit.u(0, 0, -np.pi / 16, 0)
        circuit.u(np.pi / 2, 0, 15 * np.pi / 16, 0)

        depth = 4
        basis_gates = ["h", "t", "tdg", "s", "sdg", "z"]
        gate_approx_library = generate_basic_approximations(basis_gates=basis_gates, depth=depth)

        skd = SolovayKitaev(recursion_degree=2, basic_approximations=gate_approx_library)
        discretized = skd(circuit)

        included_gates = set(discretized.count_ops().keys())
        self.assertTrue(included_gates.issubset(basis_gates))

    def test_load_from_file(self):
        """Test loading basic approximations from a file works.

        Regression test of Qiskit/qiskit#12576.
        """
        filename = "approximations.bin"

        with tempfile.TemporaryDirectory() as tmp_dir:
            fullpath = os.path.join(tmp_dir, filename)

            # dump approximations to file
            gate_approx_library = generate_basic_approximations(
                basis_gates=["h", "s", "sdg"], depth=3, filename=fullpath
            )

            # circuit to decompose and reference decomp
            circuit = QuantumCircuit(1)
            circuit.rx(0.8, 0)

            # Run SK pass using gate_approx_library
            reference = SolovayKitaev(basic_approximations=gate_approx_library)(circuit)

            # Run SK pass using stored basis_approximations
            discretized = SolovayKitaev(basic_approximations=fullpath)(circuit)

        # Check that both flows produce the same result
        self.assertEqual(discretized, reference)

    def test_load_legacy_format(self):
        """Test loading basic approximations from a legacy npy format."""
        filename = "approximations.npy"

        # build a minimal set of basic approximations, which is stored in the format
        # {gate_labels: (matrix_so3, phase)}
        approximations = {}
        gates = {gate.name: gate.to_matrix() for gate in [HGate(), TGate()]}
        for label1, matrix1 in gates.items():
            # single gates
            su2, phase = _convert_u2_to_su2(matrix1)
            so3 = _convert_su2_to_so3(su2)
            approximations[label1] = (so3, phase)

            for label2, matrix2 in gates.items():
                # gate products
                su2, phase = _convert_u2_to_su2(matrix2.dot(matrix1))
                so3 = _convert_su2_to_so3(su2)
                approximations[f"{label1}  {label2}"] = so3

        with tempfile.TemporaryDirectory() as tmp_dir:
            fullpath = os.path.join(tmp_dir, filename)

            # store in the legacy NPY format
            np.save(fullpath, approximations)

            # circuit to decompose and reference decomp
            circuit = QuantumCircuit(1)
            circuit.rx(0.8, 0)

            # Run SK pass using gate_approx_library
            reference = SolovayKitaev(basic_approximations=approximations)(circuit)

            # Run SK pass using stored basis_approximations
            discretized = SolovayKitaev(basic_approximations=fullpath)(circuit)

        # Check that both flows produce the same result
        self.assertEqual(discretized, reference)

    def test_measure(self):
        """Test the Solovay-Kitaev transpiler pass on circuits with measure operators."""
        qc = QuantumCircuit(1, 1)
        qc.x(0)
        qc.measure(0, 0)
        transpiled = self.default_sk(qc)
        self.assertEqual(set(transpiled.count_ops()), {"h", "t", "measure"})

    def test_barrier(self):
        """Test the Solovay-Kitaev transpiler pass on circuits with barriers."""
        qc = QuantumCircuit(1)
        qc.x(0)
        qc.barrier(0)
        transpiled = self.default_sk(qc)
        self.assertEqual(set(transpiled.count_ops()), {"h", "t", "barrier"})

    def test_parameterized_gates(self):
        """Test the Solovay-Kitaev transpiler pass on circuits with parameterized gates."""
        qc = QuantumCircuit(1)
        qc.x(0)
        qc.rz(Parameter("t"), 0)
        transpiled = self.default_sk(qc)
        self.assertEqual(set(transpiled.count_ops()), {"h", "t", "rz"})

    def test_control_flow_if(self):
        """Test the Solovay-Kitaev transpiler pass on circuits with control flow ops"""
        qr = QuantumRegister(1)
        cr = ClassicalRegister(1)
        qc = QuantumCircuit(qr, cr)

        with qc.if_test((cr[0], 0)) as else_:
            qc.y(0)
        with else_:
            qc.z(0)
        transpiled = self.default_sk(qc)

        # check that we still have an if-else block and all the operations within
        # have been recursively synthesized
        self.assertEqual(transpiled[0].name, "if_else")
        for block in transpiled[0].operation.blocks:
            self.assertTrue(set(block.count_ops()).issubset({"h", "t", "tdg"}))

    def test_no_to_matrix(self):
        """Test the Solovay-Kitaev transpiler pass ignores gates without to_matrix."""
        qc = QuantumCircuit(1)
        qc.initialize("0")

        transpiled = self.default_sk(qc)
        self.assertEqual(set(transpiled.count_ops()), {"initialize"})

    def test_y_gate(self):
        """Test the Solovay-Kitaev decomposition on the circuit with a Y-gate (see issue #9552)."""
        circuit = QuantumCircuit(1)
        circuit.y(0)

        transpiled = self.default_sk(circuit)
        diff = _trace_distance(circuit, transpiled)
        self.assertLess(diff, 1e-6)

    @data(["unitary"], ["rz"])
    def test_sk_synth_gates_to_basis(self, synth_gates):
        """Verify two qubit unitaries are synthesized to match basis gates."""
        unitary = QuantumCircuit(1)
        unitary.h(0)
        unitary_op = Operator(unitary)

        qc = QuantumCircuit(1)
        qc.unitary(unitary_op, 0)
        qc.rz(0.1, 0)
        dag = circuit_to_dag(qc)

        basis_gates = ["h", "t", "tdg"]
        out = UnitarySynthesis(basis_gates=basis_gates, synth_gates=synth_gates, method="sk").run(
            dag
        )
        self.assertTrue(set(out.count_ops()).isdisjoint(synth_gates))

    def test_generate_basis_approximation_gates(self):
        """Test the basis approximation generation works for all supported gates.

        Regression test of Qiskit/qiskit-terra#9585.
        """
        basis = ["i", "x", "y", "z", "h", "t", "tdg", "s", "sdg", "sx", "sxdg"]
        approx = generate_basic_approximations(basis, depth=2)

        # This mainly checks that there are no errors in the generation (like
        # in computing the inverse as described in #9585), so a simple check is enough.
        self.assertGreater(len(approx), len(basis))


@ddt
class TestSolovayKitaevDecomposition(QiskitTestCase):
    """Tests for the underlying SK decomposition class."""

    def setUp(self):
        super().setUp()
        self.default_sk = SolovayKitaevDecomposition()

    @combine(use_matrix=[True, False], gate_cls=[RXGate, RYGate, RZGate])
    def test_approximation(self, use_matrix, gate_cls):
        """Test the approximation works."""
        recursion = 4
        for angle in np.pi / np.arange(1, 10, 2):
            with self.subTest(angle=angle):
                gate = gate_cls(angle)
                if use_matrix:
                    gate = gate.to_matrix()
                synth = self.default_sk.run(gate, recursion)

                error = np.linalg.norm(Operator(gate).data - Operator(synth).data)
                self.assertLess(error, 2e-3)

    def test_query_basic_approximation(self):
        """Test finding the basic approximations for some gates."""
        # exactly representable for depth of 12
        for gate_cls in [RXGate, RYGate, RZGate]:
            for angle in [np.pi, np.pi / 2, np.pi / 4]:
                with self.subTest(gate_cls=gate_cls, angle=angle):
                    gate = gate_cls(angle)
                    synth = self.default_sk.query_basic_approximation(gate)
                    error = np.linalg.norm(Operator(gate).data - Operator(synth).data)
                    self.assertLess(error, 1e-10)

        # approximately representable
        for gate_cls in [RXGate, RYGate, RZGate]:
            for angle in [np.pi / 8, np.pi / 10, np.pi / 20]:
                with self.subTest(gate_cls=gate_cls, angle=angle):
                    gate = gate_cls(angle)
                    synth = self.default_sk.query_basic_approximation(gate)
                    error = np.linalg.norm(Operator(gate).data - Operator(synth).data)
                    self.assertLess(error, 0.5)

    @data(True, False)
    def test_basis_gates(self, use_str):
        """Test specifying the basis gates by string or gate works."""
        if use_str:
            basis_gates = ["h", "s", "sdg"]
        else:
            basis_gates = [HGate(), SGate(), SdgGate()]

        sk = SolovayKitaevDecomposition(basis_gates=basis_gates, depth=3)

        synth = sk.run(RXGate(np.pi / 2), 3)

        expected = QuantumCircuit(1, global_phase=7 * np.pi / 4)
        expected.h(0)
        expected.s(0)
        expected.h(0)

        self.assertEqual(synth, expected)


def _convert_u2_to_su2(u2_matrix: np.ndarray) -> tuple[np.ndarray, float]:
    """Convert a U(2) matrix to SU(2) by adding a global phase."""
    z = 1 / np.sqrt(np.linalg.det(u2_matrix))
    su2_matrix = z * u2_matrix
    phase = np.arctan2(np.imag(z), np.real(z))

    return su2_matrix, phase


def _convert_su2_to_so3(matrix: np.ndarray) -> np.ndarray:
    """Computes SO(3)-matrix from input SU(2)-matrix.

    Args:
        matrix: The SU(2)-matrix for which a corresponding SO(3)-matrix needs to be computed.

    Returns:
        The SO(3)-matrix corresponding to ``matrix``.

    Raises:
        ValueError: if ``matrix`` is not an SU(2)-matrix.
    """
    matrix = matrix.astype(complex)
    a = np.real(matrix[0][0])
    b = np.imag(matrix[0][0])
    c = -np.real(matrix[0][1])
    d = -np.imag(matrix[0][1])
    rotation = np.array(
        [
            [a**2 - b**2 - c**2 + d**2, 2 * a * b + 2 * c * d, -2 * a * c + 2 * b * d],
            [-2 * a * b + 2 * c * d, a**2 - b**2 + c**2 - d**2, 2 * a * d + 2 * b * c],
            [2 * a * c + 2 * b * d, 2 * b * c - 2 * a * d, a**2 + b**2 - c**2 - d**2],
        ],
        dtype=float,
    )
    return rotation


if __name__ == "__main__":
    unittest.main()
